"""Unit tests for Span classes.
"""

from itertools import combinations
from unittest import TestCase

import numpy
import pytest

from cogent3 import DNA, make_seq
from cogent3.core.location import (
    FeatureMap,
    IndelMap,
    LostSpan,
    Span,
    TerminalPadding,
    gap_coords_to_map,
)


class SpanTests(TestCase):
    """Tests of the Span object."""

    def setUp(self):
        """Define some standard Spans"""
        self.empty = Span(0, 0)
        self.full = Span(35, 30)  # will convert to (30, 35) internally
        self.overlapping = Span(32, 36)
        self.inside = Span(31, 32)
        self.before = Span(25, 30)
        self.after = Span(35, 40)
        self.reverse = Span(30, 35, reverse=True)
        self.spans_zero = Span(-5, 5)

    def test_init(self):
        """Span object should init with start, end, and Length"""
        s = Span(0)
        self.assertEqual(s.start, 0)
        self.assertEqual(s.end, 1)
        self.assertEqual(s.reverse, False)
        # to get an empty interval, must specify start and end explicitly
        t = Span(0, 0)
        self.assertEqual(t.start, 0)
        self.assertEqual(t.end, 0)
        self.assertEqual(t.reverse, False)
        # should be able to specify direction also
        u = Span(5, 15, reverse=True)
        self.assertEqual(u.start, 5)
        self.assertEqual(u.end, 15)
        self.assertEqual(u.reverse, True)
        # should be able to init from another span
        v = Span(u)
        self.assertEqual(v.start, 5)
        self.assertEqual(v.end, 15)
        self.assertEqual(v.reverse, True)

    def test_contains(self):
        """Span object contains its start but not its end"""
        self.assertNotIn(0, self.empty)
        self.assertIn(30, self.full)
        self.assertIn(34, self.full)
        self.assertNotIn(35, self.full)
        self.assertIn(self.inside, self.full)
        self.assertNotIn(self.overlapping, self.full)
        self.assertIn(0, self.spans_zero)
        self.assertIn(-5, self.spans_zero)
        self.assertNotIn(5, self.spans_zero)

    def test_overlaps(self):
        """Span objects should be able to overlap points or spans"""
        self.assertTrue(self.full.overlaps(self.overlapping))
        self.assertFalse(self.full.overlaps(self.before))
        self.assertFalse(self.before.overlaps(self.overlapping))
        self.assertFalse(self.full.overlaps(self.after))
        self.assertFalse(self.after.overlaps(self.before))
        self.assertTrue(self.full.overlaps(self.inside))
        self.assertTrue(self.spans_zero.overlaps(self.empty))
        self.assertTrue(self.empty.overlaps(self.spans_zero))

    def test_reverses(self):
        """Span.reverses should change direction"""
        self.assertFalse(self.empty.reverse)
        self.empty.reverses()
        self.assertTrue(self.empty.reverse)
        self.empty.reverses()
        self.assertFalse(self.empty.reverse)
        self.assertTrue(self.reverse.reverse)
        self.reverse.reverses()
        self.assertFalse(self.reverse.reverse)

    def test_iter(self):
        """Span iter should loop through (integer) contents"""
        self.assertEqual(list(iter(self.empty)), [])
        self.assertEqual(list(iter(self.full)), [30, 31, 32, 33, 34])
        self.assertEqual(
            list(iter(self.spans_zero)), [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4]
        )
        self.assertEqual(list(iter(self.inside)), [31])
        self.assertEqual(list(self.reverse), [34, 33, 32, 31, 30])

    def test_str(self):
        """Span str should print start, stop, reverse"""
        self.assertEqual(str(self.empty), "(0,0,False)")
        self.assertEqual(str(self.full), "(30,35,False)")
        self.assertEqual(str(self.reverse), "(30,35,True)")

    def test_len(self):
        """Span len should return difference between start and end"""
        self.assertEqual(len(self.empty), 0)
        self.assertEqual(len(self.full), 5)
        self.assertEqual(len(self.inside), 1)
        self.assertEqual(len(self.spans_zero), 10)

    def test_cmp(self):
        """Span cmp should support sort by 1st/2nd index and direction"""
        s, e, f, r, i, o = (
            self.spans_zero,
            self.empty,
            self.full,
            self.reverse,
            self.inside,
            self.overlapping,
        )

        n = Span(30, 36)

        expected_order = [s, e, f, r, n, i, o]
        first = expected_order[:]
        first.sort()
        second = [r, o, f, s, e, i, n]
        second.sort()
        for i, j in zip(first, second):
            self.assertIs(i, j)

        for i, j in zip(first, expected_order):
            self.assertIs(i, j)

    def test_sort(self):
        """Span should support sort by 1st/2nd index and direction"""
        s, e, f, r, i, o = (
            self.spans_zero,
            self.empty,
            self.full,
            self.reverse,
            self.inside,
            self.overlapping,
        )

        expected_order = [s, e]
        first = expected_order[:]
        first.sort()

        for i, j in zip(first, expected_order):
            self.assertIs(i, j)

    def test_starts_before(self):
        """Span starts_before should match hand-calculated results"""
        e, f = self.empty, self.full
        self.assertTrue(e.starts_before(f))
        self.assertFalse(f.starts_before(e))
        self.assertTrue(e.starts_before(1))
        self.assertTrue(e.starts_before(1000))
        self.assertFalse(e.starts_before(0))
        self.assertFalse(e.starts_before(-1))
        self.assertFalse(f.starts_before(30))
        self.assertTrue(f.starts_before(31))
        self.assertTrue(f.starts_before(1000))

    def test_starts_after(self):
        """Span starts_after should match hand-calculated results"""
        e, f = self.empty, self.full
        self.assertFalse(e.starts_after(f))
        self.assertTrue(f.starts_after(e))
        self.assertFalse(e.starts_after(1))
        self.assertFalse(e.starts_after(1000))
        self.assertFalse(e.starts_after(0))
        self.assertTrue(e.starts_after(-1))
        self.assertTrue(f.starts_after(29))
        self.assertFalse(f.starts_after(30))
        self.assertFalse(f.starts_after(31))
        self.assertFalse(f.starts_after(1000))

    def test_startsAt(self):
        """Span startsAt should return True if input matches"""
        e, f = self.empty, self.full
        s = Span(30, 1000)
        self.assertTrue(e.starts_at(0))
        self.assertTrue(f.starts_at(30))
        self.assertTrue(s.starts_at(30))
        self.assertTrue(f.starts_at(s))
        self.assertTrue(s.starts_at(f))
        self.assertFalse(e.starts_at(f))
        self.assertFalse(e.starts_at(-1))
        self.assertFalse(e.starts_at(1))
        self.assertFalse(f.starts_at(29))

    def test_startsInside(self):
        """Span startsInside should return True if input starts inside span"""
        e, f, i, o = self.empty, self.full, self.inside, self.overlapping
        self.assertFalse(e.starts_inside(0))
        self.assertFalse(f.starts_inside(30))
        self.assertFalse(e.starts_inside(f))
        self.assertTrue(i.starts_inside(f))
        self.assertFalse(f.starts_inside(i))
        self.assertTrue(o.starts_inside(f))
        self.assertFalse(o.ends_inside(i))

    def test_endsBefore(self):
        """Span endsBefore should match hand-calculated results"""
        e, f = self.empty, self.full
        self.assertTrue(e.ends_before(f))
        self.assertFalse(f.ends_before(e))
        self.assertTrue(e.ends_before(1))
        self.assertTrue(e.ends_before(1000))
        self.assertFalse(e.ends_before(0))
        self.assertFalse(e.ends_before(-1))
        self.assertFalse(f.ends_before(30))
        self.assertFalse(f.ends_before(31))
        self.assertTrue(f.ends_before(1000))

    def test_endsAfter(self):
        """Span endsAfter should match hand-calculated results"""
        e, f = self.empty, self.full
        self.assertFalse(e.ends_after(f))
        self.assertTrue(f.ends_after(e))
        self.assertFalse(e.ends_after(1))
        self.assertFalse(e.ends_after(1000))
        self.assertFalse(e.ends_after(0))
        self.assertTrue(e.ends_after(-1))
        self.assertTrue(f.ends_after(29))
        self.assertTrue(f.ends_after(30))
        self.assertTrue(f.ends_after(34))
        self.assertFalse(f.ends_after(35))
        self.assertFalse(f.ends_after(1000))

    def test_endsAt(self):
        """Span endsAt should return True if input matches"""
        e, f = self.empty, self.full
        s = Span(30, 1000)
        t = Span(-100, 35)
        self.assertTrue(e.ends_at(0))
        self.assertTrue(f.ends_at(35))
        self.assertTrue(s.ends_at(1000))
        self.assertFalse(f.ends_at(s))
        self.assertFalse(s.ends_at(f))
        self.assertTrue(f.ends_at(t))
        self.assertTrue(t.ends_at(f))

    def test_ends_inside(self):
        """Span ends_inside should return True if input ends inside span"""
        e, f, i, o = self.empty, self.full, self.inside, self.overlapping
        self.assertFalse(e.ends_inside(0))
        self.assertFalse(f.ends_inside(30))
        self.assertFalse(f.ends_inside(34))
        self.assertFalse(f.ends_inside(35))
        self.assertFalse(e.ends_inside(f))
        self.assertTrue(i.ends_inside(f))
        self.assertFalse(f.ends_inside(i))
        self.assertFalse(o.ends_inside(f))
        self.assertFalse(o.ends_inside(i))
        self.assertTrue(e.ends_inside(Span(-1, 1)))
        self.assertTrue(e.ends_inside(Span(0, 1)))
        self.assertFalse(e.ends_inside(Span(-1, 0)))


class MapTests(TestCase):
    """tests of the Map class"""

    def test_get_gap_coords(self):
        """returns gap start and lengths"""
        m, _ = DNA.make_seq("-AC--GT-TTA--").parse_out_gaps()
        got = m.get_gap_coordinates()
        self.assertEqual(dict(got), {0: 1, 2: 2, 4: 1, 7: 2})


def test_map_plus_position():
    # seq is 9 long
    # plus coords  012345678
    # +slice         **
    # plus seq     AAACCCTGG

    orig = FeatureMap.from_locations(locations=[(0, 9)], parent_length=9)
    assert orig.absolute_position(2) == 2
    assert orig.absolute_position(6) == 6

    assert orig.relative_position(2) == 2
    assert orig.relative_position(6) == 6

    subseq = orig[2:4]
    assert subseq.absolute_position(0) == 2
    assert subseq.absolute_position(4) == 6

    assert subseq.relative_position(2) == 0
    assert subseq.relative_position(6) == 4

    # minus coords 012345678
    # rel coord      01234
    # -slice         *****
    # minus seq    CCAGGGTTT
    # plus coords  876543210
    rc = orig.nucleic_reversed()
    rcsubseq = rc[2:7]
    abs_coord = rcsubseq.absolute_position(0)


def test_indel_map_useful_complete():
    im = IndelMap.from_spans(spans=[LostSpan(3)], parent_length=0)
    assert not im.useful
    assert not im.complete
    assert len(im) == 3


def test_map_nucleic_reversed():
    expect = [(0, 9)]
    # seq is 9 long
    # plus coords  012345678
    # plus seq     AAACCCTGG

    orig = FeatureMap.from_locations(locations=[(0, 9)], parent_length=9)
    # minus coords 012345678
    # rel coord      01234
    # -slice         *****
    # minus seq    CCAGGGTTT
    # plus coords  876543210
    rc = orig.nucleic_reversed()
    coords = rc.get_coordinates()
    assert coords == expect


@pytest.mark.parametrize("cls", (FeatureMap, IndelMap))
def test_coordinate(cls):
    # coordinates are for ungapped segments in underlying sequence
    #                   01   2 345
    seq = DNA.make_seq("AC---G-TAA--")
    m, _ = seq.parse_out_gaps()
    m = cls.from_spans(spans=tuple(m.spans), parent_length=m.parent_length)
    got = m.get_coordinates()
    assert got == [(0, 2), (2, 3), (3, 6)]


@pytest.mark.parametrize("cls", (FeatureMap, IndelMap))
def test_gap_coordinate(cls):
    seq = DNA.make_seq("AC---G-TAA--")
    m, _ = seq.parse_out_gaps()
    m = cls.from_spans(spans=tuple(m.spans), parent_length=m.parent_length)
    got = m.get_gap_coordinates()
    assert [tuple(c) for c in got] == [(2, 3), (3, 1), (6, 2)]


def test_gaps():
    # returns spans corresponding to position on "aligned" seq of gaps
    #                   000000000011
    #                   012345678901
    seq = DNA.make_seq("AC---G-TAA--")
    m, _ = seq.parse_out_gaps()
    m = FeatureMap.from_spans(spans=tuple(m.spans), parent_length=m.parent_length)
    got = [(g.start, g.end) for g in tuple(m.gaps().spans)]
    assert got == [(2, 5), (6, 7), (10, 12)]


@pytest.mark.parametrize("cls", (IndelMap, FeatureMap))
def test_nongap(cls):
    # returns spans corresponding to position on "aligned" seq of nongaps
    #                   000000000011
    #                   012345678901
    seq = DNA.make_seq("AC---G-TAA--")
    m, _ = seq.parse_out_gaps()
    m = cls.from_spans(spans=tuple(m.spans), parent_length=m.parent_length)

    got = [(s.start, s.end) for s in m.nongap()]
    assert got == [(0, 2), (5, 6), (7, 10)]


@pytest.mark.parametrize("cls", (IndelMap, FeatureMap))
def test_nongap_startswith(cls):
    # returns spans corresponding to position on "aligned" seq of nongaps
    #                   012345678
    seq = DNA.make_seq("--G-TAA--")
    m, _ = seq.parse_out_gaps()
    m = cls.from_spans(spans=tuple(m.spans), parent_length=m.parent_length)

    got = [(s.start, s.end) for s in m.nongap()]
    assert got == [(2, 3), (4, 7)]


@pytest.mark.parametrize("cls", (IndelMap, FeatureMap))
def test_nongap_not_endswith(cls):
    # returns spans corresponding to position on "aligned" seq of nongaps
    #                   0123456
    seq = DNA.make_seq("--G-TAA")
    m, _ = seq.parse_out_gaps()
    m = cls.from_spans(spans=tuple(m.spans), parent_length=m.parent_length)

    got = [(s.start, s.end) for s in m.nongap()]
    assert got == [(2, 3), (4, 7)]


def test_spans_gen():
    # returns spans corresponding to position on "aligned" seq of nongaps
    #                   000000000011
    #                   012345678901
    seq = DNA.make_seq("AC---G-TAA--")
    expect = [Span(0, 2), LostSpan(3), Span(2, 3), LostSpan(1), Span(3, 6), LostSpan(2)]
    m, _ = seq.parse_out_gaps()
    gap_data = numpy.array([(2, 3), (3, 1), (6, 2)])
    pos, lengths = gap_data.T
    im = IndelMap(gap_pos=pos, gap_lengths=lengths, parent_length=m.parent_length)
    got = list(im.spans)
    assert got == expect


def test_spans_gap_start():
    # returns spans corresponding to position on "aligned" seq of nongaps
    #                   000000000011
    #                   012345678901
    seq = DNA.make_seq("---TAA")
    expect = [LostSpan(3), Span(0, 3)]
    im, _ = seq.parse_out_gaps()
    got = list(im.spans)
    assert got == expect


def test_spans_gen_nogaps():
    # returns spans corresponding to position on "aligned" seq of nongaps
    #                   000000000011
    #                   012345678901
    seq = DNA.make_seq("ACGTAA")
    m, _ = seq.parse_out_gaps()
    spans = list(m.spans)
    assert len(spans) == 1
    assert len(spans[0]) == len(seq)


def test_round_trip_rich_dict():
    seq = DNA.make_seq("AC---G-TAA--")
    im, _ = seq.parse_out_gaps()
    rd = im.to_rich_dict()
    got = IndelMap.from_rich_dict(rd)
    assert im is not got
    assert got.to_rich_dict() == im.to_rich_dict()


def test_terminal_unknown():
    # seq  idx   01   2 345  6
    #           -AC---G-TAA--
    # aligned seq length is 13
    gap_data = numpy.array([[0, 1], [2, 3], [3, 1], [6, 2]])
    gap_pos, gap_lengths = gap_data.T

    m = IndelMap(
        gap_pos=gap_pos.copy(), gap_lengths=gap_lengths.copy(), parent_length=6
    )
    # not unknown, by default
    m_spans = tuple(m.spans)
    assert m_spans[0].lost and not isinstance(m_spans[0], TerminalPadding)
    # use the constructor arg
    m = IndelMap(
        gap_pos=gap_pos.copy(),
        gap_lengths=gap_lengths.copy(),
        parent_length=6,
        termini_unknown=True,
    )
    m_spans = tuple(m.spans)
    assert isinstance(m_spans[0], TerminalPadding)
    assert isinstance(m_spans[-1], TerminalPadding)
    assert m_spans[2].lost and not isinstance(m_spans[1], TerminalPadding)
    assert m_spans[4].lost and not isinstance(m_spans[2], TerminalPadding)

    # use the method
    m = IndelMap(
        gap_pos=gap_pos.copy(), gap_lengths=gap_lengths.copy(), parent_length=6
    ).with_termini_unknown()
    m_spans = tuple(m.spans)
    assert isinstance(m_spans[0], TerminalPadding)
    assert isinstance(m_spans[-1], TerminalPadding)
    # middle gap is not terminal, so...
    assert not isinstance(m_spans[2], TerminalPadding)

    # no gaps, no effect
    # use the constructor arg
    empty = numpy.array([], dtype=int)
    m = IndelMap(
        gap_pos=empty.copy(),
        gap_lengths=empty.copy(),
        parent_length=6,
        termini_unknown=True,
    )
    m_spans = tuple(m.spans)
    assert len(m_spans) == 1 and not m_spans[0].lost
    # use the method
    m = IndelMap(
        gap_pos=empty.copy(), gap_lengths=empty.copy(), parent_length=6
    ).with_termini_unknown()
    m_spans = tuple(m.spans)
    assert len(m_spans) == 1 and not m_spans[0].lost


def test_featuremap_inverse():
    m = FeatureMap.from_locations(locations=[(0, 2), (4, 6)], parent_length=6)
    assert len(m) == 4
    mi = m.inverse()
    assert len(mi) == 6
    mi_spans = tuple(mi.spans)
    assert mi_spans[1].lost and len(mi_spans[1]) == 2
    # invert the inversion, should give us back the original
    re_inv = mi.inverse()
    expect = m.to_rich_dict()
    got = re_inv.to_rich_dict()
    assert got == expect


def test_indelmap_from_aligned_segments():
    locations = [(0, 2), (4, 6)]
    im = IndelMap.from_aligned_segments(locations=locations, aligned_length=6)
    assert len(im) == 6
    expected_length = sum(e - s for s, e in locations)
    assert im.parent_length == expected_length
    im_spans = tuple(im.spans)
    assert im_spans[1].lost and len(im_spans[1]) == 2


@pytest.mark.parametrize("swap", (False, True))
def test_indelmap_inconsistent_input(swap):
    a = numpy.array([0, 1], dtype=int)
    b = numpy.array([3], dtype=int)
    a, b = (b, a) if swap else (a, b)
    with pytest.raises(ValueError):
        IndelMap(gap_pos=a, gap_lengths=b, parent_length=10)

    with pytest.raises(ValueError):
        IndelMap(gap_pos=a, cum_gap_lengths=b, parent_length=10)


def test_indelmap_from_aligned_segments2():
    locations = [(0, 5), (7, 12), (14, 21), (24, 27)]
    im = IndelMap.from_aligned_segments(locations=locations, aligned_length=27)
    expected_length = sum(e - s for s, e in locations)
    assert im.parent_length == expected_length


def test_indelmap_merge():
    im1 = IndelMap(
        gap_pos=numpy.array([], dtype=int),
        gap_lengths=numpy.array([], dtype=int),
        parent_length=4,
    )
    im2 = IndelMap(
        gap_pos=numpy.array([0], dtype=int),
        gap_lengths=numpy.array([2], dtype=int),
        parent_length=4,
    )

    assert len(im1) == 4
    assert im1.parent_length == 4
    inv = im1.merge_maps(im2)
    assert inv.parent_length == 4
    assert len(inv) == 6

    fmap1 = im1.to_feature_map()
    fmap2 = im2.to_feature_map()
    got = fmap2.inverse()[fmap1.inverse()].inverse()
    assert inv.get_gap_coordinates() == [
        list(coords) for coords in got.get_gap_coordinates()
    ]


def test_indelmap_merge_parent_length():
    im1 = IndelMap(
        gap_pos=numpy.array([], dtype=int),
        gap_lengths=numpy.array([], dtype=int),
        parent_length=4,
    )
    im2 = IndelMap(
        gap_pos=numpy.array([0], dtype=int),
        gap_lengths=numpy.array([2], dtype=int),
        parent_length=4,
    )
    # providing a value for parent_length overrides standard
    ov = im1.merge_maps(im2, parent_length=20)
    assert ov.parent_length != im2.parent_length
    assert ov.parent_length == 20


def test_map_indexed():
    m = FeatureMap.from_locations(locations=[(0, 2), (4, 6)], parent_length=6).inverse()
    indexed = m[2]
    assert len(indexed) == 1


def test_compare_map_indexed():
    from cogent3.core.alignment import Aligned

    raw_seq = "--AC-GTAA--"
    im, seq = DNA.make_seq(raw_seq).parse_out_gaps()
    ia = Aligned(im, seq)
    length = len(raw_seq)
    got = [str(ia[i]) for i in range(length)]
    expect = list("--AC-GTAA--")
    assert got == expect


@pytest.mark.parametrize("slice_it", (True, False))
def test_feature_map_zeroed(slice_it):
    spans = [LostSpan(2), Span(2, 4), LostSpan(2), Span(4, 8), LostSpan(2)]
    kwargs = dict(spans=spans, parent_length=6)

    fm = FeatureMap(**kwargs)
    if slice_it:
        fm = fm[3:6]

    fm_zeroed = fm.zeroed()
    assert fm.start > 0
    assert fm_zeroed.start == 0


def test_indelmap_to_feature_map():
    spans = [LostSpan(2), Span(2, 4), LostSpan(2), Span(4, 8), LostSpan(2)]
    kwargs = dict(spans=spans, parent_length=8)
    im = IndelMap.from_spans(**kwargs)
    mm = im.to_feature_map()
    assert mm.get_coordinates() == im.get_coordinates()


@pytest.mark.parametrize("raw", ("--AC--GGGG--", "A-A-A", "-A-AA----A"))
def test_indelmap_nucleic_reversed(raw):
    from cogent3.core.alignment import Aligned

    plus = DNA.make_seq(raw)
    minus = plus.rc()
    plus_imap, _ = DNA.make_seq(raw).parse_out_gaps()
    minus_imap, minus_seq = minus.parse_out_gaps()
    got = plus_imap.nucleic_reversed()
    assert got.get_coordinates() == minus_imap.get_coordinates()
    assert (got.gap_pos == minus_imap.gap_pos).all()
    assert (got.cum_gap_lengths == minus_imap.cum_gap_lengths).all()
    assert got.parent_length == minus_imap.parent_length
    assert str(Aligned(got, minus_seq)) == str(minus)


def test_get_coords():
    """get_coordinates should return raw coordinates matching input"""
    spans = [(0, 9), (20, 32)]
    fmap = FeatureMap.from_locations(locations=spans, parent_length=100)
    coords = fmap.get_coordinates()
    assert coords == spans


def test_indel_map_get_coords():
    """get_coordinates should return raw coordinates matching input"""
    imap = IndelMap(
        gap_pos=numpy.array([9]), gap_lengths=numpy.array([10]), parent_length=20
    )
    locations = [(0, 9), (9, 20)]
    coords = imap.get_coordinates()
    assert coords == locations


def test_indel_map_get_gap_coords():
    gap_data = numpy.array([(2, 3), (3, 1), (6, 2)])
    gap_pos, lengths = gap_data[:, 0], gap_data[:, 1]
    m = IndelMap(gap_pos=gap_pos, gap_lengths=lengths, parent_length=10)
    got = m.get_gap_coordinates()
    assert got == gap_data.tolist()


@pytest.mark.parametrize(
    "coords", ([(0, 3), (7, 11)], [(0, 3)], [(2, 4), (6, 10)], [(4, 6)])
)
def test_indelmap_joined_segments(coords):
    raw = "--AC--GGGG--"
    expect, _ = DNA.make_seq("".join(raw[s:e] for s, e in coords)).parse_out_gaps()
    imap, _ = DNA.make_seq(raw).parse_out_gaps()
    got = imap.joined_segments(coords)
    assert got.gap_pos.tolist() == expect.gap_pos.tolist()
    assert got.cum_gap_lengths.tolist() == expect.cum_gap_lengths.tolist()


def test_indelmap_slice_terminating():
    raw = "-CB-A--"
    start, end = 4, 6
    expect, _ = DNA.make_seq(raw[start:end]).parse_out_gaps()
    imap, _ = DNA.make_seq(raw).parse_out_gaps()
    got = imap[start:end]
    assert got.gap_pos.tolist() == expect.gap_pos.tolist()
    assert got.cum_gap_lengths.tolist() == expect.cum_gap_lengths.tolist()


def test_indelmap_slice_zero():
    raw = "-CB-A--"
    start, end = 0, 0
    expect, s = DNA.make_seq(raw[start:end]).parse_out_gaps()
    imap, _ = DNA.make_seq(raw).parse_out_gaps()
    got = imap[start:end]
    assert got.parent_length == len(s)
    assert got.gap_pos.tolist() == expect.gap_pos.tolist()
    assert got.cum_gap_lengths.tolist() == expect.cum_gap_lengths.tolist()


def test_indelmap_invalid_slice_range():
    imap = IndelMap(
        gap_pos=numpy.array([10], dtype=int),
        gap_lengths=numpy.array([2], dtype=int),
        parent_length=10,
    )
    with pytest.raises(IndexError):
        imap[-100]

    with pytest.raises(IndexError):
        imap[-100:]

    with pytest.raises(IndexError):
        imap[:-99]


def test_indelmap_get_indices_errors():
    imap = IndelMap(
        gap_pos=numpy.array([10], dtype=int),
        gap_lengths=numpy.array([2], dtype=int),
        parent_length=10,
    )
    with pytest.raises(IndexError):
        imap.get_align_index(-1000)


def test_indelmap_slice_innner_gap():
    start, end = 4, 6
    raw = "--AC--GGGG--"
    expect, _ = DNA.make_seq(raw[start:end]).parse_out_gaps()
    imap, _ = DNA.make_seq(raw).parse_out_gaps()
    imap = imap[start:end]
    assert imap.gap_pos.tolist() == expect.gap_pos.tolist()


def test_indelmap_slice_cum_length():
    start, end = 7, 11
    raw = "--AC--GGGG--"
    expect, _ = DNA.make_seq(raw[start:end]).parse_out_gaps()
    imap, _ = DNA.make_seq(raw).parse_out_gaps()
    imap = imap[start:end]
    assert imap.gap_pos.tolist() == expect.gap_pos.tolist()
    assert imap.cum_gap_lengths.tolist() == expect.cum_gap_lengths.tolist()


def test_get_coords_invalid_order():
    """get_coordinates should return raw coordinates matching input"""

    # should work for reversed Maps too
    spans = [(32, 20), (9, 0)]
    with pytest.raises(ValueError):
        FeatureMap.from_locations(locations=spans, parent_length=100)


def test_indelmap_slice_gap_into_seq():
    pos, lengths = numpy.array([[3, 1], [7, 1]]).T
    gaps = IndelMap(gap_pos=pos, gap_lengths=lengths, parent_length=7)
    # example seq
    # 012345678
    # 012 3456
    # ACA-CGAC-
    # slicing from aligned coords 3-5, gives '-C'
    got = gaps[3:5]
    assert got.gap_pos.tolist() == [0]
    assert got.cum_gap_lengths.tolist() == [1]


def test_indelmap_slice_one_gap():
    pos, lengths = numpy.array([[3, 6]]).T
    gaps = IndelMap(gap_pos=pos, gap_lengths=lengths, parent_length=24)
    # example seq
    #           1
    # 01234567890
    # 012      3456789
    # GTG------GTAGAAGTTCCAAATAATGAA
    # slicing from aligned coords 3-5, gives 'TG------G'
    # in sliced seq, gap at 2
    got = gaps[1:10]
    assert got.gap_pos.tolist() == [2]
    assert got.cum_gap_lengths.tolist() == [6]


@pytest.mark.parametrize(
    "data",
    (
        "AB---CD--EF",
        "---ABCD--EF",
        "ABCD---EF--",
        "-----ABCDEF",
        "ABCDEF-----",
        "-ABCDEF----",
        "-A-B-C-D-EF",
        "A-B-C-D-EF-",
    ),
)
@pytest.mark.parametrize("index", range(4, 6))  # the ungapped sequence is 6 long
def test_gapped_convert_seq2aln(data, index):
    # converting a sequence index to alignment index
    ungapped = data.replace("-", "")
    seq = make_seq(data, moltype="text")
    gaps, _ = seq.parse_out_gaps()
    idx = gaps.get_align_index(index)
    assert data[idx] == ungapped[index]


@pytest.mark.parametrize(
    "data",
    (
        "AB---CD--EF",
        "---ABCD--EF",
        "ABCD---EF--",
        "-----ABCDEF",
        "ABCDEF-----",
        "-ABCDEF----",
        "-A-B-C-D-EF",
        "A-B-C-D-EF-",
    ),
)
@pytest.mark.parametrize(
    "start,end", list(combinations(range(6), 2))
)  # the ungapped sequence is 6 long
def test_indelmap_align_index_slice_stop(data, start, end):
    # converting a sequence index to alignment index
    ungapped = data.replace("-", "")
    seq = make_seq(data, moltype="text")
    gaps, _ = seq.parse_out_gaps()
    stop = gaps.get_align_index(end, slice_stop=True)
    assert data[stop - 1] == ungapped[end - 1]


@pytest.mark.parametrize(
    "data",
    (
        "AB---CD--EF",
        "---ABCD--EF",
        "ABCD---EF--",
        "-----ABCDEF",
        "ABCDEF-----",
        "-ABCDEF----",
        "-A-B-C-D-EF",
        "A-B-C-D-EF-",
    ),
)
@pytest.mark.parametrize("index", range(6))  # the ungapped sequence is 6 long
def test_gapped_convert_seq2aln2seq(data, index):
    # round tripping seq to alignment to seq indices
    gaps, _ = make_seq(data, moltype="text").parse_out_gaps()
    align_index = gaps.get_align_index(index)
    got = gaps.get_seq_index(align_index)
    assert got == index


@pytest.mark.parametrize("expect", range(10))
def test_indelmap_get_seq_index_negative(expect):
    parent_length = 10
    gap_pos = [0]
    gap_lengths = [3]
    imap = IndelMap(
        gap_pos=numpy.array(gap_pos),
        gap_lengths=numpy.array(gap_lengths),
        parent_length=parent_length,
    )
    neg_index = expect - parent_length
    got = imap.get_seq_index(neg_index)
    assert got == expect


@pytest.mark.parametrize("expect", range(10))
def test_indelmap_get_align_index_negative(expect):
    parent_length = 10
    gap_pos = [0]
    gap_lengths = [3]
    gap_length = gap_lengths[0]
    imap = IndelMap(
        gap_pos=numpy.array(gap_pos),
        gap_lengths=numpy.array(gap_lengths),
        parent_length=parent_length,
    )
    neg_index = expect + gap_length - len(imap)
    got = imap.get_seq_index(neg_index)
    assert got == expect


@pytest.mark.parametrize(
    "data",
    (
        "AB--CDE-FG",
        "--ABC-DEFG",
        "AB--CDE-FG--",
        "ABCDE--FG---",
        "-----ABCDEFG",
        "-A-B-C-D-E-F-G-",
    ),
)
@pytest.mark.parametrize("seq_index", range(7))
def test_gapped_convert_aln2seq_nongap_char(data, seq_index):
    # test alignment indexes when aligned position is NOT a gap
    ungapped = "ABCDEFG"
    align_index = data.find(ungapped[seq_index])
    gaps, _ = make_seq(data, moltype="text").parse_out_gaps()
    idx = gaps.get_seq_index(align_index)
    assert idx == seq_index


def _find_nth_gap_index(data: str, n: int) -> int:
    num = -1
    for i, c in enumerate(data):
        if c == "-":
            num += 1
        if num == n:
            return i
    raise ValueError(f"{data=}, {n=}")


def _get_expected_seqindex(data: str, align_index: int) -> int:
    # compute the expected seqindex
    refseq = data.replace("-", "")
    got = data[align_index:].lstrip("-")
    return refseq.find(got[0]) if got else len(refseq)


@pytest.mark.parametrize(
    "data",
    (
        "AB-----CDE-F--G",
        "----ABC-DEFG---",
        "AB--CDE-FG-----",
        "ABCDE--FG------",
        "--------ABCDEFG",
        "-A-B-C-D-E-F-G-",
    ),
)
@pytest.mark.parametrize("gap_number", range(8))
def test_gapped_convert_aln2seq_gapchar(data, gap_number):
    # test alignment indexes when aligned position IS a gap
    # in this case we expect the position of the next non-gap
    # to be the result
    # find the alignment index corresponding to the
    align_index = _find_nth_gap_index(data, gap_number)
    assert data[align_index] == "-", (data, gap_number)
    # find nearest non-gap
    seq_index = _get_expected_seqindex(data, align_index)
    gaps, _ = make_seq(data, moltype="text").parse_out_gaps()
    idx = gaps.get_seq_index(align_index)
    assert idx == seq_index


def test_gapped_convert_aln2seq_invalid():
    gaps, _ = make_seq("AC--GTA-TG", moltype="dna").parse_out_gaps()
    with pytest.raises(IndexError):
        # absolute value of negative indices must be < seq length
        gaps.get_seq_index(-100)


@pytest.mark.parametrize(
    "invalid_slice",
    (slice(None, None, -1), slice(None, None, 2)),
)
def test_gap_pos_invalid_slice(invalid_slice):
    pos, lengths = numpy.array([[1, 3]], dtype=numpy.int32).T
    gp = IndelMap(gap_pos=pos, gap_lengths=lengths, parent_length=20)
    with pytest.raises(NotImplementedError):
        _ = gp[invalid_slice]


@pytest.mark.parametrize(
    "aslice",
    (
        slice(3, 7),
        slice(20, None),
    ),
)
def test_no_gaps_in_slice(aslice):
    # aligned length is 25
    seq_length = 20
    gap_length = 5
    pos, lengths = numpy.array([[10, gap_length]], dtype=numpy.int32).T
    gp = IndelMap(gap_pos=pos, gap_lengths=lengths, parent_length=seq_length)
    got = gp[aslice]
    assert not got.num_gaps
    start = aslice.start or 0
    stop = aslice.stop or (seq_length + gap_length)
    assert len(got) == stop - start


def test_len_gapped():
    seq_length = 20
    gap_length = 5
    pos, lengths = numpy.array([[10, gap_length]], dtype=numpy.int32).T
    gp = IndelMap(gap_pos=pos, gap_lengths=lengths, parent_length=seq_length)
    assert len(gp) == (seq_length + gap_length)


def test_all_gaps_in_slice():
    # slicing GapPositions
    # sample seq 1
    data = "AC--GTA-TG"
    gp, _ = make_seq(data, moltype="dna").parse_out_gaps()
    sl = slice(1, 9)

    got = gp[sl]
    expect_gaps, _ = make_seq(data[sl], moltype="dna").parse_out_gaps()
    assert got.get_gap_coordinates() == expect_gaps.get_gap_coordinates()
    assert got.parent_length == 5


@pytest.mark.parametrize(
    "data",
    (
        "----GTA-TG",
        "AC--GTA---",
        "AC--GTA-TG",
        "A-C-G-T-A-",
        "-A-C-G-T-A",
        "ACGTAACGTA",
        "----------",
    ),
)
@pytest.mark.parametrize(
    "aslice",
    [slice(i, j) for i, j in combinations(range(10), 2)],
)
def test_variant_slices(data, aslice):
    gaps, _ = make_seq(data, moltype="dna").parse_out_gaps()
    got = gaps[aslice]

    expect_gaps, expect_seq = make_seq(data[aslice], moltype="dna").parse_out_gaps()
    assert got.parent_length == len(expect_seq)
    assert (got.gap_pos == expect_gaps.gap_pos).all()
    assert (got.cum_gap_lengths == expect_gaps.cum_gap_lengths).all()


def test_indelmap_joined():
    pos = numpy.array([0, 1])
    cum_len = numpy.array([1, 5])
    imap = IndelMap(gap_pos=pos, cum_gap_lengths=cum_len, parent_length=5)
    fmap = FeatureMap.from_locations(locations=[(0, 1), (2, 5)], parent_length=10)
    coords = fmap.get_coordinates()
    got = imap.joined_segments(coords)
    assert got.num_gaps == 1
    assert got.gap_pos[0] == 0
    assert got.cum_gap_lengths[0] == (1 + 5 - 2)


def test_indel_map_sliced_with_negative():
    imap = IndelMap(
        gap_pos=numpy.array([0]), cum_gap_lengths=numpy.array([1]), parent_length=14
    )
    got = imap[:-3]
    assert got.parent_length == 14 - 3


def test_indelmap_roundtrip_json():
    from cogent3.util.deserialise import deserialise_object

    imap = IndelMap(
        gap_pos=numpy.array([0, 9]),
        cum_gap_lengths=numpy.array([1, 3]),
        parent_length=14,
    )
    got = deserialise_object(imap.to_json())
    assert (got.gap_pos == imap.gap_pos).all()
    assert (got.cum_gap_lengths == imap.cum_gap_lengths).all()
    assert got.parent_length == imap.parent_length


def test_featuremap_roundtrip_json():
    from cogent3.util.deserialise import deserialise_object

    fmap = FeatureMap.from_locations(
        locations=[[0, 9], [20, 23]],
        parent_length=140,
    )
    got = deserialise_object(fmap.to_json())
    coords = fmap.get_coordinates()
    assert coords == [(0, 9), (20, 23)]
    assert got.parent_length == fmap.parent_length == 140


@pytest.mark.parametrize(
    "error_type,locations", ((ValueError, ((-2, 2),)), (RuntimeError, ((20, 25),)))
)
def test_invalid_locations(error_type, locations):
    with pytest.raises(error_type):
        FeatureMap.from_locations(locations=locations, parent_length=10)


def test_gap_coords_to_map():
    """construct a Map from coordinates of gap alone"""
    gap_coords = {0: 1, 2: 2, 4: 1, 7: 2}
    seqlen = 70
    got = gap_coords_to_map(gap_coords, seqlen)
    assert len(got) == seqlen + sum(gap_coords.values())

    gap_coords = {5: 2, 17: 3, 10: 2}
    seqlen = 20
    got = gap_coords_to_map(gap_coords, seqlen)
    assert len(got) == sum(gap_coords.values()) + seqlen

    # roundtrip from Map.get_gap_coordinates()
    assert dict(got.get_gap_coordinates()) == gap_coords

    # and no gaps
    m, seq = DNA.make_seq("ACGTTTA").parse_out_gaps()
    got = gap_coords_to_map({}, len(seq))
    assert len(got) == len(m)
    assert got.get_coordinates() == m.get_coordinates()

    # and gaps outside sequence
    with pytest.raises(ValueError):
        gap_coords_to_map({20: 1}, len(seq))


@pytest.mark.parametrize(
    "seq, expected",
    [
        ("ACGTACGT", []),
        ("----ACGTACGT----", [(0, 4), (12, 16)]),
        ("ACGT----ACGT", [4, 8]),
        ("----", [[0, 4]]),
        ("----ACGT", [[0, 4]]),
        ("ACGTACGT----", [(8, 12)]),
    ],
)
def test_get_gap_align_coordinates1(seq, expected):
    expected = numpy.array(expected, dtype=int)
    if not len(expected):
        expected = expected.reshape((0, 2))
    im, _ = make_seq(seq).parse_out_gaps()
    result = im.get_gap_align_coordinates()
    assert (result == expected).all(), f"{expected=}, {result=}"


@pytest.mark.parametrize(
    "seq, expected",
    [
        ("", []),
        ("A", []),
        ("-", [[0, 1]]),
    ],
)
def test_get_gap_align_coordinates_edge_cases(seq, expected):
    expected = numpy.array(expected, dtype=int)
    if not len(expected):
        expected = expected.reshape((0, 2))

    im, _ = make_seq(seq).parse_out_gaps()
    result = im.get_gap_align_coordinates()
    assert (result == expected).all(), f"{expected=}, {result=}"


def test_featuremap_add():
    spans_a = [LostSpan(2), Span(2, 4)]
    spans_b = [LostSpan(2), Span(4, 8), LostSpan(2)]
    kwargs = dict(parent_length=6)

    fm_a = FeatureMap(spans=spans_a, **kwargs)
    fm_b = FeatureMap(spans=spans_b, **kwargs)
    fm_ab = fm_a + fm_b
    assert list(fm_ab.spans) == (spans_a + spans_b)


def test_featuremap_mul():
    spans = [LostSpan(2), Span(2, 4)]
    fm = FeatureMap(spans=spans, parent_length=6)
    fm_3 = fm * 3
    assert list(fm_3.spans) == [sp * 3 for sp in spans]
    assert fm_3.parent_length == 6 * 3


def test_featuremap_div():
    spans = [LostSpan(3), Span(3, 6)]
    fm_3 = FeatureMap(spans=spans, parent_length=6)
    fm_1 = fm_3 / 3
    assert list(fm_1.spans) == [sp / 3 for sp in spans]
    assert fm_1.parent_length == 6 / 3


def test_indelmap_make_seq_feature_map():
    #           1
    # 01234567890
    # AC--GTA-TAA
    # 01    234 567
    im = IndelMap(
        gap_pos=numpy.array([2, 5], dtype=int),
        gap_lengths=numpy.array([2, 1], dtype=int),
        parent_length=8,
    )
    orig_spans = [Span(1, 5)]
    align_map = FeatureMap(spans=orig_spans, parent_length=11)
    spans = [Span(1, 3)]
    expect = FeatureMap(spans=spans, parent_length=8)
    got = im.make_seq_feature_map(align_map)
    assert got.get_coordinates() == expect.get_coordinates()
    assert got.parent_length == expect.parent_length

    # ignoring lost spans
    align_map = FeatureMap(spans=orig_spans + [LostSpan(4)], parent_length=11)
    got = im.make_seq_feature_map(align_map)
    assert got.get_coordinates() == expect.get_coordinates()
    assert got.parent_length == expect.parent_length



from numba import njit


# turn off code coverage as njit-ted code not accessible to coverage


# fills in a diversity matrix from sequences of integers
@njit(cache=True)
def fill_diversity_matrix(matrix, seq1, seq2):  # pragma: no cover
    """fills the diversity matrix for valid positions.

    Assumes the provided sequences have been converted to indices with
    invalid characters being negative numbers (use get_moltype_index_array
    plus seq_to_indices)."""

    for i in range(len(seq1)):
        if seq1[i] < 0 or seq2[i] < 0:
            continue
        matrix[seq1[i], seq2[i]] += 1.0



from unittest import TestCase

from cogent3.recalculation.definition import CalcDefn, ParamDefn
from cogent3.recalculation.scope import (
    InvalidDimensionError,
    InvalidScopeError,
)


class RecalculationTest(TestCase):
    def test_recalculation(self):
        def add(*args):
            return sum(args)

        top = CalcDefn(add)(ParamDefn("A"), ParamDefn("B"))
        pc = top.make_likelihood_function()
        f = pc.make_calculator()

        self.assertEqual(f.get_value_array(), [1.0, 1.0])
        self.assertEqual(f([3.0, 4.25]), 7.25)
        self.assertEqual(f.change([(1, 4.5)]), 7.5)
        self.assertEqual(f.get_value_array(), [3.0, 4.5])

        # Now with scopes.  We will set up the calculation
        # result = (Ax+Bx) + (Ay+By) + (Az+Bz)

        # A and B will remain distinct parameters, but x,y and z are merely scopes - ie:
        # it may be the case that Ax = Ay = Az, and that may simplify the calculation, but
        # we will never even notice if Ax = Bx.
        # Each scope dimension (here there is just one, 'category') must be collapsed away
        # at some point towards the end of the calculation if the calculation is to produce
        # a scalar result.  Here this is done with the select_from_dimension method.

        a = ParamDefn("A", dimensions=["category"])
        b = ParamDefn("B", dimensions=["category"])
        mid = CalcDefn(add, name="mid")(a, b)
        top = CalcDefn(add)(
            mid.select_from_dimension("category", "x"),
            mid.select_from_dimension("category", "y"),
            mid.select_from_dimension("category", "z"),
        )

        # or equivalently:
        # top = CalcDefn(add, *mid.acrossDimension('category', ['x', 'y', 'z']))

        pc = top.make_likelihood_function()
        f = pc.make_calculator()

        self.assertEqual(str(f.get_value_array()), "[1.0, 1.0]")

        # There are still only 2 inputs because the default scope
        # is global, ie: Ax == Ay == Az.  If we allow A to be
        # different in the x,y and z categories and set their
        # initial values to 2.0:

        pc.assign_all("A", value=2.0, independent=True)
        f = pc.make_calculator()

        self.assertEqual(str(f.get_value_array()), "[1.0, 2.0, 2.0, 2.0]")

        # Now we have A local and B still global, so the calculation is
        # (Ax+B) + (Ay+B) + (Az+B) with the input parameters being
        # [B, Ax, Ay, Az], so:

        self.assertEqual(f([1.0, 2.0, 2.0, 2.0]), 9.0)
        self.assertEqual(f([0.25, 2.0, 2.0, 2.0]), 6.75)

        # Constants do not appear in the optimisable inputs.
        # Set one of the 3 A values to be a constant and there
        # will be one fewer optimisable parameters:

        pc.assign_all("A", scope_spec={"category": "z"}, const=True)
        f = pc.make_calculator()

        self.assertEqual(str(f.get_value_array()), "[1.0, 2.0, 2.0]")

        # The parameter controller should catch cases where the specified scope
        # does not exist:

        with self.assertRaises(InvalidScopeError):
            pc.assign_all("A", scope_spec={"category": "nosuch"})
        with self.assertRaises(InvalidDimensionError):
            pc.assign_all("A", scope_spec={"nonsuch": "nosuch"})

        # It is complicated guesswork matching the parameters you expect with positions in
        # the value array, let alone remembering whether or not they are presented to the
        # optimiser as logs, so .get_value_array(), .change() and .__call__() should only be
        # used by optimisers.  For other purposes there is an alternative, human friendly
        # interface:

        pc.update_from_calculator(f)
        self.assertEqual(pc.get_param_value("A", category="x"), 2, 0)
        self.assertEqual(pc.get_param_value("B", category=["x", "y"]), 1.0)

        # Despite the name, .get_param_value can get the value from any step in the
        # calculation, so long as it has a unique name.

        self.assertEqual(pc.get_param_value("mid", category="x"), 3.0)

        # For bulk retrieval of parameter values by parameter name and scope name there is
        # the .get_param_value_dict() method:

        vals = pc.get_param_value_dict(["category"])
        self.assertEqual(vals["A"]["x"], 2.0)

        # Here is a function that is more like a likelihood function, in that it has a
        # maximum:

        def curve(x, y):
            return 0 - (x**2 + y**2)

        top = CalcDefn(curve)(ParamDefn("X"), ParamDefn("Y"))
        pc = top.make_likelihood_function()
        f = pc.make_calculator()

        # Now ask it to find the maximum.  It is a simple function with only one local
        # maximum so local optimisation should be enough:

        f.optimise(local=True, show_progress=False)
        pc.update_from_calculator(f)

        # There were two parameters, X and Y, and at the maximum they should both be 0.0:

        self.assertEqual(pc.get_param_value("Y"), 0.0)
        self.assertEqual(pc.get_param_value("X"), 0.0)

        # Because this function has a maximum it is possible to ask it for a confidence
        # interval around a parameter, ie: how far from 0.0 can we move x before f(x,y)
        # falls bellow f(X,Y)-dropoff:

        self.assertEqual(
            pc.get_param_interval("X", dropoff=4, xtol=0.0), (-2.0, 0.0, 2.0)
        )

        # We test the ability to omit xtol. Due to precision issues we convert the returned value to a string.

        self.assertTrue(
            "-2.0, 0.0, 2.0"
            == "%.1f, %.1f, %.1f" % pc.get_param_interval("X", dropoff=4)
        )

        # And finally intervals can be calculated in bulk by passing a dropoff value to
        # .get_param_value_dict():

        self.assertEqual(
            pc.get_param_value_dict([], dropoff=4, xtol=0.0)["X"], (-2.0, 0.0, 2.0)
        )

        # For likelihood functions it is more convenient to provide 'p' rather than
        # 'dropoff', dropoff = chi2.isf(1, p) / 2.0.  Also in general you won't need ultra precise answers,
        # so don't use 'xtol=0.0', that's just to make the doctest work.
        pc.graphviz()



"""Translations of functions from Release 2.3 of the Cephes Math Library,
(c) Stephen L. Moshier 1984, 1995.
"""

from numpy import exp, floor, log, sin, sqrt

from cogent3.util import warning as c3warn


log_epsilon = 1e-6  # for threshold in log/exp close to 1
# For IEEE arithmetic (IBMPC):
MACHEP = 1.11022302462515654042e-16  # 2**-53
MAXLOG = 7.09782712893383996843e2  # log(2**1024)
MINLOG = -7.08396418532264106224e2  # log(2**-1022)
MAXNUM = 1.7976931348623158e308  # 2**1024

PI = 3.14159265358979323846  # pi
PIO2 = 1.57079632679489661923  # pi/2
PIO4 = 7.85398163397448309616e-1  # pi/4
SQRT2 = 1.41421356237309504880  # sqrt(2)
SQRTH = 7.07106781186547524401e-1  # sqrt(2)/2
LOG2E = 1.4426950408889634073599  # 1/log(2)
SQ2OPI = 7.9788456080286535587989e-1  # sqrt( 2/pi )
LOGE2 = 6.93147180559945309417e-1  # log(2)
LOGSQ2 = 3.46573590279972654709e-1  # log(2)/2
THPIO4 = 2.35619449019234492885  # 3*pi/4
TWOOPI = 6.36619772367581343075535e-1  # 2/pi

ROUND_ERROR = 1e-14  # fp rounding error: causes some tests to fail
# will round to 0 if smaller in magnitude than this


def fix_rounding_error(x):
    """If x is almost in the range 0-1, fixes it.

    Specifically, if x is between -ROUND_ERROR and 0, returns 0.
    If x is between 1 and 1+ROUND_ERROR, returns 1.
    """
    if -ROUND_ERROR < x < 0:
        return 0
    elif 1 < x < 1 + ROUND_ERROR:
        return 1
    else:
        return x


def log_one_minus(x):
    """Returns natural log of (1-x). Useful for probability calculations."""
    if abs(x) < log_epsilon:
        return -x
    else:
        return log(1 - x)


def one_minus_exp(x):
    """Returns 1-exp(x). Useful for probability calculations."""
    if abs(x) < log_epsilon:
        return -x
    else:
        return 1 - exp(x)


@c3warn.deprecated_callable(
    version="2024.9", reason="use scipy.special.perm() instead", is_discontinued=True
)
def permutations(n, k):  # pragma: no cover
    """Returns the number of ways of choosing k items from n, in order.

    Defined as n!/(n-k)!.
    """
    # Validation: k must be be between 0 and n (inclusive), and n must be >=0.
    if k > n:
        raise IndexError(f"Can't choose {k} items from {n}")
    elif k < 0:
        raise IndexError("Can't choose negative number of items")
    elif n < 0:
        raise IndexError("Can't choose from negative number of items")
    if min(n, k) < 20 and isinstance(n, int) and isinstance(k, int):
        return permutations_exact(n, k)
    else:
        return exp(ln_permutations(n, k))


@c3warn.deprecated_callable(
    version="2024.9", reason="use scipy.special.perm() instead", is_discontinued=True
)
def permutations_exact(n, k):  # pragma: no cover
    """Calculates permutations by integer division.

    Preferred method for small permutations, but slow on larger ones.

    Note: no error checking (expects to be called through permutations())
    """
    product = 1
    for i in range(n - k + 1, n + 1):
        product *= i
    return product


@c3warn.deprecated_callable(
    version="2024.9", reason="use scipy.special.perm() instead", is_discontinued=True
)
def ln_permutations(n, k):  # pragma: no cover
    """Calculates permutations by difference in log of gamma function.

    Preferred method for large permutations, but slow on smaller ones.

    Note: no error checking (expects to be called through permutations())
    """
    return lgam(n + 1) - lgam(n - k + 1)


@c3warn.deprecated_callable(
    version="2024.9", reason="Use scipy.special.comb() instead", is_discontinued=True
)
def combinations(n, k):  # pragma: no cover
    """Returns the number of ways of choosing k items from n, in order.

    Defined as n!/(k!(n-k)!).
    """
    # Validation: k must be be between 0 and n (inclusive), and n must be >=0.
    if k > n:
        raise IndexError(f"Can't choose {k} items from {n}")
    elif k < 0:
        raise IndexError("Can't choose negative number of items")
    elif n < 0:
        raise IndexError("Can't choose from negative number of items")
    # if min(n, k) < 20:
    if min(n, k) < 20 and isinstance(n, int) and isinstance(k, int):
        return combinations_exact(n, k)
    else:
        return exp(ln_combinations(n, k))


@c3warn.deprecated_callable(
    version="2024.9", reason="Use scipy.special.comb() instead", is_discontinued=True
)
def combinations_exact(n, k):  # pragma: no cover
    """Calculates combinations by integer division.

    Preferred method for small combinations, but slow on larger ones.

    Note: no error checking (expects to be called through combinations())
    """
    # permutations(n, k) = permutations(n, n-k), so reduce computation by
    # figuring out which requires calculation of fewer terms.
    if k > (n - k):
        larger = k
        smaller = n - k
    else:
        larger = n - k
        smaller = k

    product = 1
    # compute n!/(n-larger)! by multiplying terms from n to (n-larger+1)
    for i in range(larger + 1, n + 1):
        product *= i

    # divide by (smaller)! by multiplying terms from 2 to smaller
    for i in range(2, smaller + 1):  # no need to divide by 1...
        product /= i  # ok to use integer division: should always be factor

    return product


@c3warn.deprecated_callable(
    version="2024.9", reason="Use scipy.special.comb() instead", is_discontinued=True
)
def ln_combinations(n, k):  # pragma: no cover
    """Calculates combinations by difference in log of gamma function.

    Preferred method for large combinations, but slow on smaller ones.

    Note: no error checking (expects to be called through combinations())
    """
    return lgam(n + 1) - lgam(k + 1) - lgam(n - k + 1)


def ln_binomial(successes, trials, prob):
    """Returns the natural log of the binomial distribution.

    successes: number of successes
    trials: number of trials
    prob: probability of success

    Works for int and float values. Approximated by the gamma function.

    Note: no error checking (expects to be called through binomial_exact())
    """
    from scipy.special import loggamma

    prob = fix_rounding_error(prob)
    ln_comb = (
        loggamma(trials + 1)
        - loggamma(successes + 1)
        - loggamma(trials - successes + 1)
    )
    return ln_comb + successes * log(prob) + (trials - successes) * log(1.0 - prob)


# Translations of functions from Cephes Math Library, by Stephen L. Moshier


def polevl(x, coef):
    """evaluates a polynomial y = C_0 + C_1x + C_2x^2 + ... + C_Nx^N

    Coefficients are stored in reverse order, i.e. coef[0] = C_N
    """
    result = 0
    for c in coef:
        result = result * x + c
    return result


# Coefficients for zdist follow:
ZP = [
    2.46196981473530512524e-10,
    5.64189564831068821977e-1,
    7.46321056442269912687e0,
    4.86371970985681366614e1,
    1.96520832956077098242e2,
    5.26445194995477358631e2,
    9.34528527171957607540e2,
    1.02755188689515710272e3,
    5.57535335369399327526e2,
]

ZQ = [
    1.0,
    1.32281951154744992508e1,
    8.67072140885989742329e1,
    3.54937778887819891062e2,
    9.75708501743205489753e2,
    1.82390916687909736289e3,
    2.24633760818710981792e3,
    1.65666309194161350182e3,
    5.57535340817727675546e2,
]

ZR = [
    5.64189583547755073984e-1,
    1.27536670759978104416e0,
    5.01905042251180477414e0,
    6.16021097993053585195e0,
    7.40974269950448939160e0,
    2.97886665372100240670e0,
]
ZS = [
    1.00000000000000000000e0,
    2.26052863220117276590e0,
    9.39603524938001434673e0,
    1.20489539808096656605e1,
    1.70814450747565897222e1,
    9.60896809063285878198e0,
    3.36907645100081516050e0,
]
ZT = [
    9.60497373987051638749e0,
    9.00260197203842689217e1,
    2.23200534594684319226e3,
    7.00332514112805075473e3,
    5.55923013010394962768e4,
]
ZU = [
    1.00000000000000000000e0,
    3.35617141647503099647e1,
    5.21357949780152679795e2,
    4.59432382970980127987e3,
    2.26290000613890934246e4,
    4.92673942608635921086e4,
]


@c3warn.deprecated_callable(
    version="2024.9",
    reason="use scipy.special.erf(x, out=None) instead",
    is_discontinued=True,
)
def erf(a):  # pragma: no cover
    """being removed"""
    if abs(a) > 1:
        return 1 - erfc(a)
    z = a * a
    return a * polevl(z, ZT) / polevl(z, ZU)


@c3warn.deprecated_callable(
    version="2024.9",
    reason="use scipy.special.erfc(z, out=None) instead",
    is_discontinued=True,
)
def erfc(a):  # pragma: no cover
    """being removed"""
    if a < 0:
        x = -a
    else:
        x = a

    if x < 1:
        return 1 - erf(a)

    z = -a * a
    if z < -MAXLOG:  # underflow
        if a < 0:
            return 2
        else:
            return 0
    z = exp(z)

    if x < 8:
        p = polevl(x, ZP)
        q = polevl(x, ZQ)
    else:
        p = polevl(x, ZR)
        q = polevl(x, ZS)

    y = z * p / q

    if a < 0:
        y = 2 - y

    if y == 0:  # underflow
        if a < 0:
            return 2
        else:
            return 0
    else:
        return y


# Coefficients for Gamma follow:
GA = [
    8.11614167470508450300e-4,
    -5.95061904284301438324e-4,
    7.93650340457716943945e-4,
    -2.77777777730099687205e-3,
    8.33333333333331927722e-2,
]

GB = [
    -1.37825152569120859100e3,
    -3.88016315134637840924e4,
    -3.31612992738871184744e5,
    -1.16237097492762307383e6,
    -1.72173700820839662146e6,
    -8.53555664245765465627e5,
]

GC = [
    1.00000000000000000000e0,
    -3.51815701436523470549e2,
    -1.70642106651881159223e4,
    -2.20528590553854454839e5,
    -1.13933444367982507207e6,
    -2.53252307177582951285e6,
    -2.01889141433532773231e6,
]

GP = [
    1.60119522476751861407e-4,
    1.19135147006586384913e-3,
    1.04213797561761569935e-2,
    4.76367800457137231464e-2,
    2.07448227648435975150e-1,
    4.94214826801497100753e-1,
    9.99999999999999996796e-1,
]

GQ = [
    -2.31581873324120129819e-5,
    5.39605580493303397842e-4,
    -4.45641913851797240494e-3,
    1.18139785222060435552e-2,
    3.58236398605498653373e-2,
    -2.34591795718243348568e-1,
    7.14304917030273074085e-2,
    1.00000000000000000320e0,
]

STIR = [
    7.87311395793093628397e-4,
    -2.29549961613378126380e-4,
    -2.68132617805781232825e-3,
    3.47222221605458667310e-3,
    8.33333333333482257126e-2,
]

MAXSTIR = 143.01608
MAXLGM = 2.556348e305
MAXGAM = 171.624376956302725
LOGPI = 1.14472988584940017414
SQTPI = 2.50662827463100050242e0
LS2PI = 0.91893853320467274178

# Generally useful constants
SQRTH = 7.07106781186547524401e-1
SQRT2 = 1.41421356237309504880
MAXLOG = 7.09782712893383996843e2
MINLOG = -7.08396418532264106224e2
MACHEP = 1.11022302462515654042e-16
PI = 3.14159265358979323846

big = 4.503599627370496e15
biginv = 2.22044604925031308085e-16


def igamc(a, x):
    """Complemented incomplete Gamma integral: see Cephes docs."""
    if x <= 0 or a <= 0:
        return 1
    if x < 1 or x < a:
        return 1 - igam(a, x)
    ax = a * log(x) - x - lgam(a)
    if ax < -MAXLOG:  # underflow
        return 0
    ax = exp(ax)
    # continued fraction
    y = 1 - a
    z = x + y + 1
    c = 0
    pkm2 = 1
    qkm2 = x
    pkm1 = x + 1
    qkm1 = z * x
    ans = pkm1 / qkm1

    while 1:
        c += 1
        y += 1
        z += 2
        yc = y * c
        pk = pkm1 * z - pkm2 * yc
        qk = qkm1 * z - qkm2 * yc
        if qk != 0:
            r = pk / qk
            t = abs((ans - r) / r)
            ans = r
        else:
            t = 1
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk
        if abs(pk) > big:
            pkm2 *= biginv
            pkm1 *= biginv
            qkm2 *= biginv
            qkm1 *= biginv
        if t <= MACHEP:
            break
    return ans * ax


def igam(a, x):
    """Left tail of incomplete gamma function: see Cephes docs for details"""
    if x <= 0 or a <= 0:
        return 0
    if x > 1 and x > a:
        return 1 - igamc(a, x)

    # Compute x**a * exp(x) / Gamma(a)

    ax = a * log(x) - x - lgam(a)
    if ax < -MAXLOG:  # underflow
        return 0.0
    ax = exp(ax)

    # power series
    r = a
    c = 1
    ans = 1
    while 1:
        r += 1
        c *= x / r
        ans += c
        if c / ans <= MACHEP:
            break

    return ans * ax / a


def lgam(x):
    """Natural log of the gamma fuction: see Cephes docs for details"""
    if x < -34:
        q = -x
        w = lgam(q)
        p = floor(q)
        if p == q:
            raise OverflowError("lgam returned infinity.")

        z = q - p
        if z > 0.5:
            p += 1
            z = p - q
        z = q * sin(PI * z)
        if z == 0:
            raise OverflowError("lgam returned infinity.")
        z = LOGPI - log(z) - w
        return z

    if x < 13:
        z = 1
        p = 0
        u = x
        while u >= 3:
            p -= 1
            u = x + p
            z *= u
        while u < 2:
            if u == 0:
                raise OverflowError("lgam returned infinity.")
            z /= u
            p += 1
            u = x + p
        if z < 0:
            z = -z
        if u == 2:
            return log(z)
        p -= 2
        x = x + p
        p = x * polevl(x, GB) / polevl(x, GC)
        return log(z) + p
    if x > MAXLGM:
        raise OverflowError("Too large a value of x in lgam.")
    q = (x - 0.5) * log(x) - x + LS2PI
    if x > 1.0e8:
        return q
    p = 1 / (x * x)
    if x >= 1000:
        q += (
            (7.9365079365079365079365e-4 * p - 2.7777777777777777777778e-3) * p
            + 0.0833333333333333333333
        ) / x
    else:
        q += polevl(p, GA) / x
    return q


def betai(aa, bb, xx):
    """Returns integral of the incomplete beta density function, from 0 to x.

    See Cephes docs for details.
    """
    if aa <= 0 or bb <= 0:
        raise ValueError("betai: a and b must both be > 0.")
    if xx == 0:
        return 0
    if xx == 1:
        return 1
    if xx < 0 or xx > 1:
        raise ValueError("betai: x must be between 0 and 1.")
    flag = 0
    if (bb * xx <= 1) and (xx <= 0.95):
        t = pseries(aa, bb, xx)
        return betai_result(t, flag)
    w = 1 - xx
    # reverse a and b if x is greater than the mean
    if xx > (aa / (aa + bb)):
        flag = 1
        a = bb
        b = aa
        xc = xx
        x = w
    else:
        a = aa
        b = bb
        xc = w
        x = xx
    if (flag == 1) and ((b * x) <= 1) and (x <= 0.95):
        t = pseries(a, b, x)
        return betai_result(t, flag)
    # choose expansion for better convergence
    y = x * (a + b - 2) - (a - 1)
    if y < 0:
        w = incbcf(a, b, x)
    else:
        w = incbd(a, b, x) / xc
    y = a * log(x)
    t = b * log(xc)
    if ((a + b) < MAXGAM) and (abs(y) < MAXLOG) and (abs(t) < MAXLOG):
        t = pow(xc, b)
        t *= pow(x, a)
        t /= a
        t *= w
        t *= Gamma(a + b) / (Gamma(a) * Gamma(b))
        return betai_result(t, flag)
    # resort to logarithms
    y += t + lgam(a + b) - lgam(a) - lgam(b)
    y += log(w / a)
    if y < MINLOG:
        t = 0
    else:
        t = exp(y)
    return betai_result(t, flag)


def betai_result(t, flag):
    if flag == 1:
        if t <= MACHEP:
            t = 1 - MACHEP
        else:
            t = 1 - t
    return t


incbet = betai  # shouldn't have renamed in first place...


def incbcf(a, b, x):
    """Incomplete beta integral, first continued fraction representation.

    See Cephes docs for details."""

    k1 = a
    k2 = a + b
    k3 = a
    k4 = a + 1
    k5 = 1
    k6 = b - 1
    k7 = k4
    k8 = a + 2

    pkm2 = 0
    qkm2 = 1
    pkm1 = 1
    qkm1 = 1
    ans = 1
    r = 1
    n = 0
    thresh = 3 * MACHEP

    while 1:
        xk = -(x * k1 * k2) / (k3 * k4)
        pk = pkm1 + pkm2 * xk
        qk = qkm1 + qkm2 * xk
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk

        xk = (x * k5 * k6) / (k7 * k8)
        pk = pkm1 + pkm2 * xk
        qk = qkm1 + qkm2 * xk
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk

        if qk != 0:
            r = pk / qk
        if r != 0:
            t = abs((ans - r) / r)
            ans = r
        else:
            t = 1
        if t < thresh:
            return ans
        k1 += 1
        k2 += 1
        k3 += 2
        k4 += 2
        k5 += 1
        k6 -= 1
        k7 += 2
        k8 += 2

        if (abs(qk) + abs(pk)) > big:
            pkm2 *= biginv
            pkm1 *= biginv
            qkm2 *= biginv
            qkm1 *= biginv
        if (abs(qk) < biginv) or (abs(pk) < biginv):
            pkm2 *= big
            pkm1 *= big
            qkm2 *= big
            qkm1 *= big
        n += 1
        if n >= 300:
            return ans


def incbd(a, b, x):
    """Incomplete beta integral, second continued fraction representation.

    See Cephes docs for details."""

    k1 = a
    k2 = b - 1.0
    k3 = a
    k4 = a + 1.0
    k5 = 1.0
    k6 = a + b
    k7 = a + 1.0
    k8 = a + 2.0

    pkm2 = 0.0
    qkm2 = 1.0
    pkm1 = 1.0
    qkm1 = 1.0
    z = x / (1.0 - x)
    ans = 1.0
    r = 1.0
    n = 0
    thresh = 3 * MACHEP

    while 1:
        xk = -(z * k1 * k2) / (k3 * k4)
        pk = pkm1 + pkm2 * xk
        qk = qkm1 + qkm2 * xk
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk

        xk = (z * k5 * k6) / (k7 * k8)
        pk = pkm1 + pkm2 * xk
        qk = qkm1 + qkm2 * xk
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk

        if qk != 0:
            r = pk / qk
        if r != 0:
            t = abs((ans - r) / r)
            ans = r
        else:
            t = 1.0
        if t < thresh:
            return ans
        k1 += 1
        k2 -= 1
        k3 += 2
        k4 += 2
        k5 += 1
        k6 += 1
        k7 += 2
        k8 += 2

        if (abs(qk) + abs(pk)) > big:
            pkm2 *= biginv
            pkm1 *= biginv
            qkm2 *= biginv
            qkm1 *= biginv

        if (abs(qk) < biginv) or (abs(pk) < biginv):
            pkm2 *= big
            pkm1 *= big
            qkm2 *= big
            qkm1 *= big
        n += 1
        if n >= 300:
            return ans


def Gamma(x):
    """Returns the gamma function, a generalization of the factorial.

    See Cephes docs for details."""
    if hasattr(x, "item"):
        # avoid issue of x being a limited precision numpy type
        # use item() method casts to the nearest Python type
        x = x.item()

    sgngam = 1
    q = abs(x)
    if q > 33:
        if x < 0:
            p = floor(q)
            if p == q:
                raise OverflowError("Bad value of x in Gamma function.")
            i = p
            if (i & 1) == 0:
                sgngam = -1
            z = q - p
            if z > 0.5:
                p += 1
                z = q - p
            z = q * sin(PI * z)
            if z == 0:
                raise OverflowError("Bad value of x in Gamma function.")
            z = abs(z)
            z = PI / (z * stirf(q))
        else:
            z = stirf(x)
        return sgngam * z
    z = 1
    while x >= 3:
        x -= 1
        z *= x
    while x < 0:
        if x > -1e9:
            return Gamma_small(x, z)
    while x < 2:
        if x < 1e-9:
            return Gamma_small(x, z)
        z /= x
        x += 1
    if x == 2:
        return float(z)
    x -= 2
    p = polevl(x, GP)
    q = polevl(x, GQ)
    return z * p / q


def Gamma_small(x, z):
    if x == 0:
        raise OverflowError("Bad value of x in Gamma function.")
    else:
        return z / ((1 + 0.5772156649015329 * x) * x)


def stirf(x):
    """Stirling's approximation for the Gamma function.

    Valid for 33 <= x <= 162.

    See Cephes docs for details.
    """
    w = 1.0 / x
    w = 1 + w * polevl(w, STIR)
    y = exp(x)
    if x > MAXSTIR:
        # avoid overflow in pow()
        v = pow(x, 0.5 * x - 0.25)
        y = v * (v / y)
    else:
        y = pow(x, x - 0.5) / y
    return SQTPI * y * w


def pseries(a, b, x):
    """Power series for incomplete beta integral.

    Use when b * x is small and x not too close to 1.

    See Cephes docs for details.
    """
    ai = 1 / a
    u = (1 - b) * x
    v = u / (a + 1)
    t1 = v
    t = u
    n = 2
    s = 0
    z = MACHEP * ai
    while abs(v) > z:
        u = (n - b) * x / n
        t *= u
        v = t / (a + n)
        s += v
        n += 1
    s += t1
    s += ai

    u = a * log(x)
    if ((a + b) < MAXGAM) and (abs(u) < MAXLOG):
        t = Gamma(a + b) / (Gamma(a) * Gamma(b))
        s = s * t * pow(x, a)
    else:
        t = lgam(a + b) - lgam(a) - lgam(b) + u + log(s)
        if t < MINLOG:
            s = 0
        else:
            s = exp(t)
    return s


def log1p(x):
    """Log for values close to 1: from Cephes math library"""
    z = 1 + x
    if (z < SQRTH) or (z > SQRT2):
        return log(z)
    z = x * x
    z = -0.5 * z + x * (z * polevl(x, LP) / polevl(x, LQ))
    return x + z


LP = [
    4.5270000862445199635215e-5,
    4.9854102823193375972212e-1,
    6.5787325942061044846969e0,
    2.9911919328553073277375e1,
    6.0949667980987787057556e1,
    5.7112963590585538103336e1,
    2.0039553499201281259648e1,
]
LQ = [
    1,
    1.5062909083469192043167e1,
    8.3047565967967209469434e1,
    2.2176239823732856465394e2,
    3.0909872225312059774938e2,
    2.1642788614495947685003e2,
    6.0118660497603843919306e1,
]


def expm1(x):
    """Something to do with exp? From Cephes."""
    if (x < -0.5) or (x > 0.5):
        return exp(x) - 1.0
    xx = x * x
    r = x * polevl(xx, EP)
    r /= polevl(xx, EQ) - r
    return r + r


EP = [
    1.2617719307481059087798e-4,
    3.0299440770744196129956e-2,
    9.9999999999999999991025e-1,
]

EQ = [
    3.0019850513866445504159e-6,
    2.5244834034968410419224e-3,
    2.2726554820815502876593e-1,
    2.0000000000000000000897e0,
]


def igami(a, y0):
    # bound the solution
    x0 = MAXNUM
    yl = 0
    x1 = 0
    yh = 1.0
    dithresh = 5.0 * MACHEP

    # handle easy cases
    if (y0 < 0.0) or (y0 > 1.0) or (a <= 0):
        raise ZeroDivisionError("y0 must be between 0 and 1; a >= 0")
    elif y0 == 0.0:
        return MAXNUM
    elif y0 == 1.0:
        return 0.0
    # approximation to inverse function
    d = 1.0 / (9.0 * a)
    y = 1.0 - d - ndtri(y0) * sqrt(d)
    x = a * y * y * y

    lgm = lgam(a)

    for i in range(10):
        # this loop is just to eliminate gotos
        while 1:
            if x > x0 or x < x1:
                break
            y = igamc(a, x)
            if y < yl or y > yh:
                break
            if y < y0:
                x0 = x
                yl = y
            else:
                x1 = x
                yh = y
            # compute the derivative of the function at this point
            d = (a - 1.0) * log(x) - x - lgm
            if d < -MAXLOG:
                break
            d = -exp(d)
            # compute the step to the next approximation of x
            d = (y - y0) / d
            if abs(d / x) < MACHEP:
                return x
            x -= d
            break

    # Resort to interval halving if Newton iteration did not converge.
    d = 0.0625
    if x0 == MAXNUM:
        if x <= 0.0:
            x = 1.0
        while x0 == MAXNUM:
            x = (1.0 + d) * x
            y = igamc(a, x)
            if y < y0:
                x0 = x
                yl = y
                break
            d += d
    d = 0.5
    dir = 0

    for i in range(400):
        x = x1 + d * (x0 - x1)
        y = igamc(a, x)
        lgm = (x0 - x1) / (x1 + x0)
        if abs(lgm) < dithresh:
            break
        lgm = (y - y0) / y0
        if abs(lgm) < dithresh:
            break
        if x <= 0.0:
            break
        if y >= y0:
            x1 = x
            yh = y
            if dir < 0:
                dir = 0
                d = 0.5
            elif dir > 1:
                d = 0.5 * d + 0.5
            else:
                d = (y0 - yl) / (yh - yl)
            dir += 1
        else:
            x0 = x
            yl = y
            if dir > 0:
                dir = 0
                d = 0.5
            elif dir < -1:
                d *= 0.5
            else:
                d = (y0 - yl) / (yh - yl)
            dir -= 1
    if x == 0.0:
        return 0
    return x


P0 = [
    -5.99633501014107895267e1,
    9.80010754185999661536e1,
    -5.66762857469070293439e1,
    1.39312609387279679503e1,
    -1.23916583867381258016e0,
]

Q0 = [
    1.00000000000000000000e0,
    1.95448858338141759834e0,
    4.67627912898881538453e0,
    8.63602421390890590575e1,
    -2.25462687854119370527e2,
    2.00260212380060660359e2,
    -8.20372256168333339912e1,
    1.59056225126211695515e1,
    -1.18331621121330003142e0,
]

s2pi = 2.50662827463100050242e0

P1 = [
    4.05544892305962419923e0,
    3.15251094599893866154e1,
    5.71628192246421288162e1,
    4.40805073893200834700e1,
    1.46849561928858024014e1,
    2.18663306850790267539e0,
    -1.40256079171354495875e-1,
    -3.50424626827848203418e-2,
    -8.57456785154685413611e-4,
]

Q1 = [
    1.00000000000000000000e0,
    1.57799883256466749731e1,
    4.53907635128879210584e1,
    4.13172038254672030440e1,
    1.50425385692907503408e1,
    2.50464946208309415979e0,
    -1.42182922854787788574e-1,
    -3.80806407691578277194e-2,
    -9.33259480895457427372e-4,
]

P2 = [
    3.23774891776946035970e0,
    6.91522889068984211695e0,
    3.93881025292474443415e0,
    1.33303460815807542389e0,
    2.01485389549179081538e-1,
    1.23716634817820021358e-2,
    3.01581553508235416007e-4,
    2.65806974686737550832e-6,
    6.23974539184983293730e-9,
]

Q2 = [
    1.00000000000000000000e0,
    6.02427039364742014255e0,
    3.67983563856160859403e0,
    1.37702099489081330271e0,
    2.16236993594496635890e-1,
    1.34204006088543189037e-2,
    3.28014464682127739104e-4,
    2.89247864745380683936e-6,
    6.79019408009981274425e-9,
]

exp_minus_2 = 0.13533528323661269189


def ndtri(y0):
    """Inverse normal distribution function.

    This is here and not in distributions because igami depends on it..."""
    y0 = fix_rounding_error(y0)
    # handle easy cases
    if y0 <= 0.0:
        return -MAXNUM
    elif y0 >= 1.0:
        return MAXNUM
    code = 1
    y = y0
    if y > (1.0 - exp_minus_2):
        y = 1.0 - y
        code = 0

    if y > exp_minus_2:
        y -= 0.5
        y2 = y * y
        x = y + y * (y2 * polevl(y2, P0) / polevl(y2, Q0))
        x = x * s2pi
        return x

    x = sqrt(-2.0 * log(y))
    x0 = x - log(x) / x

    z = 1.0 / x
    if x < 8.0:  # y > exp(-32) = 1.2664165549e-14
        x1 = z * polevl(z, P1) / polevl(z, Q1)
    else:
        x1 = z * polevl(z, P2) / polevl(z, Q2)
    x = x0 - x1
    if code != 0:
        x = -x
    return x


def incbi(aa, bb, yy0):
    """Incomplete beta inverse function. See Cephes for docs."""
    # handle easy cases first
    if yy0 <= 0:
        return 0.0
    elif yy0 >= 1.0:
        return 1.0

    # define inscrutable parameters
    x0 = 0.0
    yl = 0.0
    x1 = 1.0
    yh = 1.0
    nflg = 0

    if aa <= 1.0 or bb <= 1.0:
        dithresh = 1.0e-6
        rflg = 0
        a = aa
        b = bb
        y0 = yy0
        x = a / (a + b)
        y = incbet(a, b, x)
        return _incbi_ihalve(
            dithresh, rflg, nflg, a, b, x0, yl, x1, yh, y0, x, y, aa, bb, yy0
        )
    else:
        dithresh = 1.0e-4

    # approximation to inverse function
    yp = -ndtri(yy0)

    if yy0 > 0.5:
        rflg = 1
        a = bb
        b = aa
        y0 = 1.0 - yy0
        yp = -yp
    else:
        rflg = 0
        a = aa
        b = bb
        y0 = yy0

    lgm = (yp * yp - 3.0) / 6.0
    x = 2.0 / (1.0 / (2.0 * a - 1.0) + 1.0 / (2.0 * b - 1.0))
    d = yp * sqrt(x + lgm) / x - (1.0 / (2.0 * b - 1.0) - 1.0 / (2.0 * a - 1.0)) * (
        lgm + 5.0 / 6.0 - 2.0 / (3.0 * x)
    )
    d *= 2.0
    if d < MINLOG:
        x = 1.0
        return _incbi_under(rflg, x)

    x = a / (a + b * exp(d))
    y = incbet(a, b, x)
    yp = (y - y0) / y0
    if abs(yp) < 0.2:
        return _incbi_newt(
            dithresh, rflg, nflg, a, b, x0, yl, x1, yh, y0, x, y, aa, bb, yy0
        )
    else:
        return _incbi_ihalve(
            dithresh, rflg, nflg, a, b, x0, yl, x1, yh, y0, x, y, aa, bb, yy0
        )


def _incbi_done(rflg, x):
    """Final test in incbi."""
    if rflg:
        if x <= MACHEP:
            x = 1.0 - MACHEP
        else:
            x = 1.0 - x
    return x


def _incbi_under(rflg, x):
    """Underflow handler in incbi."""
    x = 0.0
    return _incbi_done(rflg, x)


class IhalveRepeat(Exception):
    pass

    # Resort to interval halving if not close enough.


def _incbi_ihalve(dithresh, rflg, nflg, a, b, x0, yl, x1, yh, y0, x, y, aa, bb, yy0):
    """Interval halving in incbi."""
    while 1:
        try:
            dir = 0
            di = 0.5
            for i in range(100):
                if i != 0:
                    x = x0 + di * (x1 - x0)
                    if x == 1.0:
                        x = 1.0 - MACHEP
                    if x == 0.0:
                        di = 0.5
                        x = x0 + di * (x1 - x0)
                        if x == 0.0:
                            return _incbi_under(rflg, x)
                    y = incbet(a, b, x)
                    yp = (x1 - x0) / (x1 + x0)
                    if abs(yp) < dithresh:
                        return _incbi_newt(
                            dithresh,
                            rflg,
                            nflg,
                            a,
                            b,
                            x0,
                            yl,
                            x1,
                            yh,
                            y0,
                            x,
                            y,
                            aa,
                            bb,
                            yy0,
                        )
                    yp = (y - y0) / y0
                    if abs(yp) < dithresh:
                        return _incbi_newt(
                            dithresh,
                            rflg,
                            nflg,
                            a,
                            b,
                            x0,
                            yl,
                            x1,
                            yh,
                            y0,
                            x,
                            y,
                            aa,
                            bb,
                            yy0,
                        )
                if y < y0:
                    x0 = x
                    yl = y
                    if dir < 0:
                        dir = 0
                        di = 0.5
                    elif dir > 3:
                        di = 1.0 - (1.0 - di) * (1.0 - di)
                    elif dir > 1:
                        di = 0.5 * di + 0.5
                    else:
                        di = (y0 - y) / (yh - yl)
                    dir += 1
                    if x0 > 0.75:
                        if rflg == 1:
                            rflg = 0
                            a = aa
                            b = bb
                            y0 = yy0
                        else:
                            rflg = 1
                            a = bb
                            b = aa
                            y0 = 1.0 - yy0
                        x = 1.0 - x
                        y = incbet(a, b, x)
                        x0 = 0.0
                        yl = 0.0
                        x1 = 1.0
                        yh = 1.0
                        raise IhalveRepeat
                else:
                    x1 = x
                    if rflg == 1 and x1 < MACHEP:
                        x = 0.0
                        return _incbi_done(rflg, x)
                    yh = y
                    if dir > 0:
                        dir = 0
                        di = 0.5
                    elif dir < -3:
                        di *= di
                    elif dir < -1:
                        di *= 0.5
                    else:
                        di = (y - y0) / (yh - yl)
                    dir -= 1
            if x0 >= 1.0:
                x = 1.0 - MACHEP
                return _incbi_done(rflg, x)
            if x <= 0.0:
                return _incbi_under(rflg, x)
        except IhalveRepeat:
            continue


def _incbi_newt(dithresh, rflg, nflg, a, b, x0, yl, x1, yh, y0, x, y, aa, bb, yy0):
    """Newton's method for incbi."""
    if nflg:
        return _incbi_done(rflg, x)
    nflg = 1
    lgm = lgam(a + b) - lgam(a) - lgam(b)

    for i in range(8):
        # Compute the function at this point.
        if i != 0:
            y = incbet(a, b, x)
        if y < yl:
            x = x0
            y = yl
        elif y > yh:
            x = x1
            y = yh
        elif y < y0:
            x0 = x
            yl = y
        else:
            x1 = x
            yh = y
        if x == 1.0 or x == 0.0:
            break
        # Compute the derivative of the function at this point.
        d = (a - 1.0) * log(x) + (b - 1.0) * log(1.0 - x) + lgm
        if d < MINLOG:
            return _incbi_done(rflg, x)
        if d > MAXLOG:
            break
        d = exp(d)
        # Compute the step to the next approximation of x.
        d = (y - y0) / d
        xt = x - d
        if xt <= x0:
            y = (x - x0) / (x1 - x0)
            xt = x0 + 0.5 * y * (x - x0)
            if xt <= 0.0:
                break
        if xt >= x1:
            y = (x1 - x) / (x1 - x0)
            xt = x1 - 0.5 * y * (x1 - x)
            if xt >= 1.0:
                break
        x = xt
        if abs(d / x) < 128.0 * MACHEP:
            return _incbi_done(rflg, x)
    # Did not converge.
    dithresh = 256.0 * MACHEP
    return _incbi_ihalve(
        dithresh, rflg, nflg, a, b, x0, yl, x1, yh, y0, x, y, aa, bb, yy0
    )



import pickle
import warnings

import pytest

from cogent3.util.warning import deprecated_args, deprecated_callable


def test_function_deprecated_args():
    @deprecated_args(
        version="a future release",
        reason="x and y are not descriptive",
        old_new=[("x", "a"), ("y", "b")],
    )
    def changed(a: int, b: int) -> int:
        return a + b

    with pytest.deprecated_call():
        expected = changed(a=5, b=3)
        got = changed(x=5, y=3)
        assert got == expected


def test_function_deprecated_args_docstring():
    @deprecated_args(
        version="a future release",
        reason="x and y are not descriptive",
        old_new=[("x", "a"), ("y", "b")],
    )
    def changed(a: int, b: int) -> int:
        """This is a test function"""
        return a + b

    assert changed.__doc__ == "This is a test function"


@pytest.mark.parametrize("kwargs", (dict(x=5, y=3), dict(a=5, y=3), dict(x=5, b=3)))
def test_function_deprecated_args_warn(kwargs):
    # Example target function to be decorated
    @deprecated_args(
        version="a future release",
        reason="x and y are not descriptive",
        old_new=[("x", "a"), ("y", "b")],
    )
    def changed(a: int, b: int) -> int:
        return a + b

    with pytest.deprecated_call():
        changed(**kwargs)


def test_function_correct_args_do_not_warn():
    @deprecated_args(
        version="a future release",
        reason="x and y are not descriptive",
        old_new=[("x", "a"), ("y", "b")],
    )
    def changed(a: int, b: int) -> int:
        return a + b

    with warnings.catch_warnings():
        # verify no warnings for correct parameters
        warnings.simplefilter("error")
        changed(a=5, b=3)


def test_function_deprecated_args_pickled():
    @deprecated_args(
        version="a future release",
        reason="x and y are not descriptive",
        old_new=[("x", "a"), ("y", "b")],
    )
    def changed(a: int, b: int) -> int:
        return a + b

    with pytest.deprecated_call():
        myfunc = changed(x=1, y=2)
        pickled_func = pickle.dumps(myfunc)
        assert isinstance(pickled_func, bytes)
        unpickled_func = pickle.loads(pickled_func)
        assert unpickled_func == changed(a=1, b=2)


class foo:
    def __init__(self):
        self.a = 0
        self.b = 0

    @deprecated_args(
        version="a future release",
        reason="x and y are not descriptive",
        old_new=[("x", "a"), ("y", "b")],
    )
    def changed(self, a: int, b: int):
        """This is a test function"""
        self.a = a
        self.b = b


def test_method_deprecated_args():
    foo_instance = foo()
    foo_instance.changed(a=5, b=3)
    assert foo_instance.a == 5
    assert foo_instance.b == 3


def test_method_deprecated_args_docstring():
    assert foo.changed.__doc__ == "This is a test function"
    foo_instance = foo()
    assert foo_instance.changed.__doc__ == "This is a test function"


@pytest.mark.parametrize("kwargs", (dict(x=5, y=3), dict(a=5, y=3), dict(x=5, b=3)))
def test_method_deprecated_args_warn(kwargs):
    with pytest.deprecated_call():
        foo().changed(**kwargs)  # pylint: disable=no-value-for-parameter


def test_method_correct_args_do_not_warn():
    with warnings.catch_warnings():
        # verify no warnings for correct parameters
        warnings.simplefilter("error")
        foo().changed(a=5, b=3)


def test_method_deprecated_args_pickled():
    foo_instance = foo()
    assert foo_instance.a == 0
    assert foo_instance.b == 0
    with pytest.deprecated_call():
        foo_instance.changed(x=1, y=2)  # pylint: disable=no-value-for-parameter
    assert foo_instance.a == 1
    assert foo_instance.b == 2

    # serialize instance with pickle
    pickled_foo = pickle.dumps(foo_instance)
    assert isinstance(pickled_foo, bytes)

    # deserialize instance with pickle
    unpickled_foo = pickle.loads(pickled_foo)
    assert isinstance(unpickled_foo, foo)
    assert unpickled_foo.a == 1
    assert unpickled_foo.b == 2
    with pytest.deprecated_call():
        unpickled_foo.changed(x=2, y=3)
    assert unpickled_foo.a == 2
    assert unpickled_foo.b == 3


class foo2:
    # noting that new_meth does not exist
    @deprecated_callable("2023.9", reason="test meth", new="new_meth")
    def old_meth(self, v):
        return v**2

    @deprecated_callable("2023.9", reason="redundant", is_discontinued=True)
    def squared(self, v):
        return v * v


@deprecated_callable("2023.9", reason="test func", new="new_func")
def old_func(v):
    return v**3


@deprecated_callable("2023.9", reason="redundant", is_discontinued=True)
def cubed(v):
    return v * v


@pytest.mark.parametrize("func", (foo2().old_meth, foo2().squared, old_func, cubed))
def test_deprecated_callable_warn(func):
    with pytest.deprecated_call():
        func(2)


@pytest.mark.parametrize("func", (cubed, old_func))
def test_method_deprecated_function_pickling(recwarn, func):
    # using pytest recwarn fixture here since the filterwarning decorator
    # does not seem to play nice with parametrize
    # serialize func with pickle
    pickled_func = pickle.dumps(func)
    assert isinstance(pickled_func, bytes)

    # deserialize with pickle
    unpickled_func = pickle.loads(pickled_func)

    assert unpickled_func(20) == func(20)


def test_method_deprecated_method_pickling(recwarn):
    # serialize instance with pickle
    instance = foo2()
    pickled = pickle.dumps(instance)
    assert isinstance(pickled, bytes)

    # deserialize with pickle
    unpickled = pickle.loads(pickled)

    # the method still works after pickling
    assert unpickled.old_meth(20) == instance.old_meth(20)

    # the method still works after pickling
    assert unpickled.squared(20) == instance.squared(20)


@pytest.mark.parametrize(
    "func,_type", ((foo2().old_meth, "method"), (cubed, "function"))
)
def test_deprecated_callable_resolves_type(recwarn, func, _type):
    func(2)
    assert _type in recwarn.list[0].message.args[0]


def test_function_deprecated_args_deprecated_callable_chained_decorators(recwarn):
    @deprecated_args("2023.6", "x is not descriptive", old_new=[("x", "a")])
    @deprecated_args("2023.6", "b is no longer required", discontinued=["b"])
    @deprecated_callable(
        "2023.6", "Improved change function", new="changed2", is_discontinued=True
    )
    def changed(a: int, b: int) -> int:
        return a + b

    got = changed(x=5, b=3)  # pylint: disable=no-value-for-parameter
    assert got == 8
    warnings = [warning.message.args[0] for warning in recwarn.list]
    assert any("argument x which will be removed" in warning for warning in warnings)
    assert any("argument b is discontinued" in warning for warning in warnings)
    assert any("function changed is discontinued" in warning for warning in warnings)


def test_class_deprecated(recwarn):
    class fooclass:
        @deprecated_callable("2023.6", "Improved change function", is_discontinued=True)
        def __init__(self): ...

    fooclass()
    warnings = [warning.message.args[0] for warning in recwarn.list]
    assert any("fooclass is discontinued" in warning for warning in warnings)



from collections import defaultdict
from typing import List, Optional, Union

from numpy import array
from numpy import random as np_random

from cogent3.core.alignment import Alignment, ArrayAlignment
from cogent3.core.genetic_code import get_code
from cogent3.core.moltype import MolType, get_moltype

from .composable import NON_COMPOSABLE, NotCompleted, define_app
from .translate import get_fourfold_degenerate_sets
from .typing import AlignedSeqsType, SeqsCollectionType, SerialisableType


# TODO need a function to filter sequences based on divergence, ala divergent
# set.

MolTypes = Union[str, MolType]


def intersection(groups):
    """returns the intersection of all groups"""
    common = set(groups.pop())
    return common.intersection(*map(set, groups))


def union(groups):
    """returns the intersection of all groups"""
    union = set(groups.pop())
    union = union.union(*map(set, groups))
    return union


@define_app(app_type=NON_COMPOSABLE)
class concat:
    """Creates a concatenated alignment from a series."""

    def __init__(
        self, join_seq: str = "", intersect: bool = True, moltype: Optional[str] = None
    ):
        """
        Parameters
        ----------
        join_seq
            splice sequences together using this string
        intersect
            result contains only sequences present in all alignments. If False,
            missings sequences will be replaced by a sequence of '?'.
        moltype
            molecular type, must be either DNA or RNA

        Examples
        --------

        Create an app to concatenate two alignments

        >>> from cogent3 import app_help, get_app, make_aligned_seqs
        >>> concat_alns = get_app("concat", moltype="dna")

        Create sample alignments with matching sequence names

        >>> aln1 = make_aligned_seqs({"s1": "AAA", "s2": "CAA", "s3": "AAA"}, moltype="dna")
        >>> aln2 = make_aligned_seqs({"s1": "GCG", "s2": "GGG", "s3": "GGT"}, moltype="dna")

        Concatenate alignments. By default, sequences without matching names in
        the corresponding alignment are omitted (intersect=True).

        >>> result = concat_alns([aln1, aln2])
        >>> print(result.to_pretty(name_order=["s1","s2","s3"]))
        s1    AAAGCG
        s2    C...G.
        s3    ....GT

        Create an app that includes missing sequences across alignments.
        Missing sequences are replaced by a sequence of "?".

        >>> concat_missing = get_app("concat", moltype="dna", intersect=False)
        >>> aln3 = make_aligned_seqs({"s4": "GCG", "s5": "GGG"}, moltype="dna")
        >>> result = concat_missing([aln1, aln3])
        >>> print(result.to_pretty(name_order=["s1","s2","s3","s4","s5"]))
        s1    AAA???
        s2    C.....
        s3    ......
        s4    ???GCG
        s5    ???GGG

        Create an app that delimits concatenated alignments with "N"

        >>> concat_delim = get_app("concat", join_seq="N", moltype="dna")
        >>> result = concat_delim([aln1, aln2])
        >>> print(result.to_pretty(name_order=["s1","s2","s3"]))
        s1    AAANGCG
        s2    C....G.
        s3    .....GT
        """
        self._name_callback = {True: intersection}.get(intersect, union)
        self._intersect = intersect
        self._moltype = moltype
        self._join_seq = join_seq

    def main(
        self, data: List[AlignedSeqsType]
    ) -> Union[SerialisableType, AlignedSeqsType]:
        """returns an alignment

        Parameters
        ----------
        data
            series of alignment instances
        """
        if not data:
            return NotCompleted("ERROR", self, message="no data")

        names = []
        for aln in data:
            if self._moltype is None:
                self._moltype = aln.moltype

            if not isinstance(aln, (ArrayAlignment, Alignment)):
                raise TypeError(f"{type(aln)} invalid for concat")
            names.append(aln.names)

        names = self._name_callback(names)
        collated = defaultdict(list)

        for aln in data:
            if self._moltype and aln.moltype != self._moltype:
                # try converting
                aln = aln.to_moltype(self._moltype)

            if self._intersect:
                seqs = aln.take_seqs(names).to_dict()
            else:
                seqs = defaultdict(lambda: "?" * len(aln))
                seqs |= aln.to_dict()

            for name in names:
                collated[name].append(seqs[name])

        combined = {n: self._join_seq.join(collated[n]) for n in names}
        if aln := ArrayAlignment(data=combined, moltype=self._moltype):
            return aln
        else:
            return NotCompleted("FAIL", self, message="result is empty")


@define_app
class omit_degenerates:
    """Excludes alignment columns with degenerate characters. Can accomodate
    reading frame."""

    def __init__(
        self,
        moltype: Optional[str] = None,
        gap_is_degen: bool = True,
        motif_length: int = 1,
    ):
        """
        Parameters
        ----------
        moltype
            molecular type, must be either DNA or RNA
        gap_is_degen
            include gap character in degenerate character set
        motif_length
            sequences split into non-overlapping tuples of this size. If a
            tuple contains a degen character at any position the entire tuple
            is excluded

        Examples
        --------
        Degenerate IUPAC base symbols represents a site position that can have
        multiple possible nucleotides. For example, "Y" represents
        pyrimidines where the site can be either "C" or "T".

        Note: In molecular evolutionary and phylogenetic analyses, the gap
        character "-" is considered to be any base "N".

        Create sample data with degenerate characters

        >>> from cogent3 import app_help, get_app, make_aligned_seqs
        >>> aln = make_aligned_seqs({"s1": "ACGA-GACG", "s2": "GATGATGYT"}, moltype="dna")

        Create an app that omits aligned columns containing a degenerate
        character from an alignment

        >>> app = get_app("omit_degenerates", moltype="dna")
        >>> result = app(aln)
        >>> print(result.to_pretty())
        s1    ACGAGAG
        s2    GATGTGT

        Create an app which omits degenerate characters, but retains gaps

        >>> app = get_app("omit_degenerates", moltype="dna", gap_is_degen=False)
        >>> result = app(aln)
        >>> print(result.to_pretty())
        s1    ACGA-GAG
        s2    GATGATGT

        Split sequences into non-overlapping tuples of length 2 and exclude
        any tuple that contains a degenerate character

        >>> app = get_app("omit_degenerates", moltype="dna", motif_length=2)
        >>> result = app(aln)
        >>> print(result.to_pretty())
        s1    ACGA
        s2    GATG

        A NotCompleted object (see https://cogent3.org/doc/app/not-completed.html)
        is returned if the moltype is not specified in the alignment or app

        >>> aln = make_aligned_seqs({"s1": "ACGA-GACG", "s2": "GATGATGYT"})
        >>> app = get_app("omit_degenerates")
        >>> result = app(aln)
        >>> result.message
        'Traceback...
        """
        if moltype:
            moltype = get_moltype(moltype)
            assert moltype.label.lower() in ("dna", "rna"), "Invalid moltype"

        self._moltype = moltype
        self._allow_gap = not gap_is_degen
        self._motif_length = motif_length

    T = Union[SerialisableType, AlignedSeqsType]

    def main(self, aln: AlignedSeqsType) -> T:
        if self._moltype and aln.moltype != self._moltype:
            # try converting
            aln = aln.to_moltype(self._moltype)

        return aln.no_degenerates(
            motif_length=self._motif_length, allow_gap=self._allow_gap
        ) or NotCompleted("FAIL", self, "all columns contained degenerates", source=aln)


@define_app
class omit_gap_pos:
    """Excludes gapped alignment columns meeting a threshold. Can accomodate
    reading frame."""

    def __init__(
        self,
        allowed_frac: float = 0.99,
        motif_length: int = 1,
        moltype: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        allowed_frac
            columns with a fraction of gap characters exceeding allowed_frac are
            excluded
        motif_length
            sequences split into non-overlapping tuples of this size.
        moltype
            molecular type, must be either DNA or RNA

        Examples
        --------

        Create a sample alignment and an app that excludes highly gapped sites.
        Sites with over 99% gaps are excluded by default.

        >>> from cogent3 import make_aligned_seqs, get_app
        >>> aln = make_aligned_seqs({"s1": "ACGA-GA-CG", "s2": "GATGATG-AT"})

        >>> app = get_app("omit_gap_pos", moltype="dna")
        >>> result = app(aln)
        >>> print(result.to_pretty(name_order=["s1", "s2"]))
        s1    ACGA-GACG
        s2    GATGATGAT

        Create an app that excludes all aligned sites with over 49% gaps.

        >>> app = get_app("omit_gap_pos", allowed_frac=0.49, moltype="dna")
        >>> result = app(aln)
        >>> print(result.to_pretty(name_order=["s1", "s2"]))
        s1    ACGAGACG
        s2    GATGTGAT

        To eliminate any codon columns (where a column is a triple of
        nucleotides) that contain a gap character, we use the motif_length
        argument.

        >>> app = get_app("omit_gap_pos", allowed_frac=0, motif_length=3, moltype="dna")
        >>> result = app(aln)
        >>> print(result.to_pretty(name_order=["s1", "s2"]))
        s1    ACG
        s2    GAT

        A NotCompleted object (see https://cogent3.org/doc/app/not-completed.html)
        is returned if all sites are excluded.

        >>> aln = make_aligned_seqs({"s1": "ACGA------", "s2": "----ATG-AT"})
        >>> app = get_app("omit_gap_pos", allowed_frac=0, motif_length=3, moltype="dna")
        >>> result = app(aln)
        >>> result.message
        'all columns exceeded gap threshold'
        """
        if moltype:
            moltype = get_moltype(moltype)
            assert moltype.label.lower() in ("dna", "rna"), "Invalid moltype"

        self._moltype = moltype
        self._allowed_frac = allowed_frac
        self._motif_length = motif_length

    T = Union[SerialisableType, AlignedSeqsType]

    def main(self, aln: AlignedSeqsType) -> T:
        if self._moltype and aln.moltype != self._moltype:
            # try converting
            aln = aln.to_moltype(self._moltype)

        return aln.omit_gap_pos(
            allowed_gap_frac=self._allowed_frac, motif_length=self._motif_length
        ) or NotCompleted(
            "FAIL", self, "all columns exceeded gap threshold", source=aln
        )


@define_app
class take_codon_positions:
    """Extracts the specified codon position(s) from an alignment."""

    def __init__(
        self,
        *positions: int,
        fourfold_degenerate: bool = False,
        gc: Union[str, int] = "Standard",
        moltype: str = "dna",
    ):
        """
        Parameters
        ----------
        positions
            either a single integer from (1, 2, 3), or additional keyword
            arguments of position numbers, e.g. 3 is third position, (1,2)
            is first and second codon position
        fourfold_degenerate
            if True, returns third positions from four-fold degenerate codons.
            Overrides positions.
        gc
            identifier for a genetic code or a genetic code instance.
            see https://cogent3.org/doc/cookbook/what_codes.html
        moltype
            molecular type, must be either DNA or RNA

        Examples
        --------

        Create a sample alignment and an app that extracts the 3rd codon
        position from an alignment.

        >>> from cogent3 import make_aligned_seqs, get_app
        >>> aln = make_aligned_seqs({"s1": "ACGACGACG", "s2": "GATGATGAT"})
        >>> take_pos3 = get_app("take_codon_positions", 3, moltype="dna")
        >>> result = take_pos3(aln)
        >>> print(result.to_pretty())
        s1    GGG
        s2    TTT

        Create an app that extracts the 1st and 2nd codon positions from an
        alignment.

        >>> take_pos12 = get_app("take_codon_positions", 1, 2, moltype="dna")
        >>> result = take_pos12(aln)
        >>> print(result.to_pretty())
        s1    ACACAC
        s2    GAGAGA

        Create a sample alignment and an app that returns the 3rd codon
        positions from four-fold degenerate codons.

        >>> aln_ff = make_aligned_seqs({"s1": "GCAAGCGTTTAT", "s2": "GCTTTTGTCAAT"})
        >>> take_fourfold = get_app("take_codon_positions", fourfold_degenerate=True, moltype="dna")
        >>> result = take_fourfold(aln_ff)
        >>> print(result.to_pretty())
        s1    AT
        s2    TC

        A NotCompleted object (see https://cogent3.org/doc/app/not-completed.html)
        is returned if all sites are excluded.

        >>> result = take_fourfold(aln)
        >>> result.message
        'Traceback ...
        """
        assert moltype is not None
        moltype = get_moltype(moltype)

        assert moltype.label.lower() in ("dna", "rna"), "Invalid moltype"

        self._moltype = moltype
        self._four_fold_degen = fourfold_degenerate
        self._fourfold_degen_sets = None

        if fourfold_degenerate:
            gc = get_code(gc)
            sets = get_fourfold_degenerate_sets(
                gc, alphabet=moltype.alphabet, as_indices=True
            )
            self._fourfold_degen_sets = sets
            self._func = self.take_fourfold_positions
            return

        assert (
            1 <= min(positions) <= 3 and 1 <= max(positions) <= 3
        ), "Invalid codon positions"

        by_index = len(positions) == 1
        if by_index:
            positions = positions[0] - 1
            self._func = self.take_codon_position
        else:
            positions = tuple(p - 1 for p in sorted(positions))
            self._func = self.take_codon_positions

        self._positions = positions

    def take_fourfold_positions(self, aln):
        if self._moltype and self._moltype != aln.moltype:
            aln = aln.to_moltype(self._moltype)

        fourfold_codon_sets = self._fourfold_degen_sets

        def ffold(x):
            x = {tuple(e) for e in list(x)}
            for codon_set in fourfold_codon_sets:
                if x <= codon_set:
                    return True
            return False

        new = aln.filtered(ffold, motif_length=3)
        return new[2::3]

    def take_codon_position(self, aln):
        if isinstance(aln, Alignment):
            indices = list(range(self._positions, len(aln), 3))
            result = aln.take_positions(indices)
        elif isinstance(aln, ArrayAlignment):
            result = aln[self._positions :: 3]
        return result

    def take_codon_positions(self, aln):
        """takes multiple positions"""
        length = len(aln)
        indices = [k for k in range(length) if k % 3 in self._positions]
        return aln.take_positions(indices)

    T = Union[SerialisableType, AlignedSeqsType]

    def main(self, aln: AlignedSeqsType) -> T:
        return self._func(aln)


@define_app
class take_named_seqs:
    """Selects named sequences from a collection."""

    def __init__(self, *names: str, negate: bool = False):
        """
        Parameters
        ----------
        *names
            series of sequence names provided as positional arguments
        negate
            if True, excludes the provided names from the result

        Examples
        --------

        Create a sample alignment and an app that returns the sequences
        matching the provided names.

        >>> from cogent3 import make_aligned_seqs, get_app
        >>> aln = make_aligned_seqs({
        ...     "s1": "GCAAGC",
        ...     "s2": "GCTTTT",
        ...     "s3": "GC--GC",
        ...     "s4": "GCAAGC"
        ... })
        >>> app = get_app("take_named_seqs", "s1", "s2")
        >>> result = app(aln)
        >>> print(result.to_pretty())
        s1    GCAAGC
        s2    ..TTTT

        Create an app that excludes sequences that match the provided names.

        >>> app_negate = get_app("take_named_seqs", "s1", "s2", negate=True)
        >>> result = app_negate(aln)
        >>> print(result.to_pretty())
        s3    GC--GC
        s4    ..AA..
        """
        self._names = names
        self._negate = negate

    T = Union[SerialisableType, SeqsCollectionType]

    def main(self, data: SeqsCollectionType) -> T:
        try:
            data = data.take_seqs(self._names, negate=self._negate)
        except KeyError:
            missing = set(self._names) - set(data.names)
            msg = f"named seq(s) {missing} not in {data.names}"
            data = NotCompleted("FALSE", self, msg, source=data)
        return data


@define_app
class take_n_seqs:
    """Selects n sequences from a collection. Chooses first n sequences, or
    selects randomly if specified."""

    def __init__(
        self,
        number: int,
        random: bool = False,
        seed: Optional[int] = None,
        fixed_choice: bool = True,
    ):
        """
        Parameters
        ----------
        number
            number of sequences to sample. If number of sequences in a collectionis < n, returns NotCompleted
            indicating a FAIL.
        random
            Whether to choose the sequences randomly.
        seed
            Seed for the numpy random number generator.
        fixed_choice
            sequence names selected from the first alignment are used for all others.

        Returns
        -------
        A new sequence collection, or NotCompleted if not insufficient sequences are in the collection.

        Examples
        --------

        Create a sample alignment and an app that returns ``number=3``
        sequences. By default, the first 3 sequences from the alignment are
        returned.

        >>> from cogent3 import make_aligned_seqs, get_app
        >>> aln = make_aligned_seqs({
        ...     "s1": "ACGT",
        ...     "s2": "ACG-",
        ...     "s3": "ACGN",
        ...     "s4": "ACGG",
        ...     "s5": "ACGG"
        ... })
        >>> app_first_n = get_app("take_n_seqs", number=3)
        >>> result = app_first_n(aln)
        >>> print(result.to_pretty())
        s1    ACGT
        s2    ...-
        s3    ...N

        Using ``random=True`` and ``number=3`` returns 3 random sequences.
        An optional ``seed`` can be provided to ensure the same sequences
        are returned each time the app is called.

        >>> app_random_n = get_app("take_n_seqs", number=3, random=True, seed=1)
        >>> result = app_random_n(aln)
        >>> print(result.to_pretty())
        s3    ACGN
        s2    ...-
        s5    ...G

        ``fixed_choice=True`` ensures the same sequences are returned when
        (randomly) sampling sequences across several alignments.

        >>> aln2 = make_aligned_seqs({
        ...     "s1": "GCGC",
        ...     "s2": "GCG-",
        ...     "s3": "GCG-",
        ...     "s4": "GCGG",
        ...     "s5": "GCGG",
        ... })
        >>> app_fixed = get_app("take_n_seqs", number=3, random=True, fixed_choice=True)
        >>> result1 = app_fixed(aln).names
        >>> result2 = app_fixed(aln2).names
        >>> assert result1 == result2

        When ``number`` exceeds the number of sequences in the alignment, returns a
        NotCompleted (see https://cogent3.org/doc/app/not-completed.html).

        >>> app_fail = get_app("take_n_seqs", number=6)
        >>> result = app_fail(aln)
        >>> result.message
        'not enough sequences'
        """
        if seed:
            np_random.seed(seed)

        self._names = None
        self._number = number
        self._random = random
        self._fixed_choice = fixed_choice

    def _set_names(self, data):
        """set the names attribute"""
        if not self._random:
            self._names = data.names[: self._number]
            return

        self._names = np_random.choice(data.names, self._number, replace=False).tolist()

    T = Union[SerialisableType, SeqsCollectionType]

    def main(self, data: SeqsCollectionType) -> T:
        """returns data with n sequences"""
        if len(data.names) < self._number:
            return NotCompleted("FALSE", self.main, "not enough sequences")

        if self._names is None or not self._fixed_choice:
            self._set_names(data)

        try:
            data = data.take_seqs(self._names)
        except KeyError:
            missing = set(self._names) - set(data.names)
            msg = f"named seq(s) {missing} not in {data.names}"
            data = NotCompleted("FALSE", self, msg, source=data)
        return data


@define_app
class min_length:
    """Filters sequence collections / alignments by length."""

    def __init__(
        self,
        length: int,
        motif_length: int = 1,
        subtract_degen: bool = True,
        moltype: Optional[MolTypes] = None,
    ):
        """
        Parameters
        ----------
        length
            only alignments with this length returned, False otherwise
        motif_length
            length is converted to modulo motif_length
        subtract_degen
            degenerate characters subtracted from sequence length calculation
        moltype
            molecular type, can be string or instance

        Examples
        --------
        Create a sample alignment. Alignments must have a moltype specified

        >>> from cogent3 import make_aligned_seqs, get_app
        >>> aln = make_aligned_seqs({"s1": "ACGGT", "s2": "AC-GT"}, moltype="dna")

        Create the app to filter sequences with a minimum length of 3 sites

        >>> app = get_app("min_length", 3)
        >>> result = app(aln)
        >>> len(result)
        5

        When all sequences are shorter than the min length, returns a
        NotCompleted (see https://cogent3.org/doc/app/not-completed.html)

        >>> app = get_app("min_length", 7)
        >>> result = app(aln)
        >>> result.message
        '4 < min_length 7'

        If the moltype of the alignment is not specified, returns a NotCompleted

        >>> aln = make_aligned_seqs({"s1": "ACGGT", "s2": "AC-GT"})
        >>> app = get_app("min_length", 3)
        >>> result = app(aln)
        >>> result.message
        'Traceback...
        """
        if motif_length > 1:
            length //= motif_length
        self._min_length = length
        self._motif_length = motif_length
        self._subtract_degen = subtract_degen
        if moltype:
            moltype = get_moltype(moltype)
        self._moltype = moltype

    T = Union[SerialisableType, SeqsCollectionType]

    def main(self, data: SeqsCollectionType) -> T:
        if self._moltype and self._moltype != data.moltype:
            data = data.to_moltype(self._moltype)

        if self._subtract_degen and not hasattr(data.alphabet, "non_degen"):
            raise ValueError(
                f"{self.__class__.__name__}(subtract_degen=True) requires DNA, RNA or PROTEIN "
                "moltype"
            )

        lengths = data.get_lengths(
            allow_gap=not self._subtract_degen,
            include_ambiguity=not self._subtract_degen,
        )
        length, _ = min((l, n) for n, l in lengths.items())

        if length < self._min_length:
            msg = f"{length} < min_length {self._min_length}"
            data = NotCompleted("FALSE", self, msg, source=data)

        return data


class _GetStart:
    choose = np_random.choice

    def __init__(self, start):
        self._start = start
        self.func = {True: self._int}.get(type(start) == int, self._rand)

    def _int(self, length):
        return self._start

    def _rand(self, length):
        return self.choose(length)

    def __call__(self, length):
        return self.func(length)


@define_app
class fixed_length:
    """Sample an alignment to a fixed length."""

    def __init__(
        self,
        length: int,
        start: int = 0,
        random: bool = False,
        seed: Optional[int] = None,
        motif_length: int = 1,
        moltype: Optional[MolTypes] = None,
    ):
        """
        Parameters
        ----------
        length
            only alignments with this length returned, False otherwise
        start
            integer starting position for truncation, or 'random' in which case
            a random start is chosen (within the possible range returning an
            alignment of the specified length). Overrides  `random`.
        random
            random positions for the corresponding tuple are chosen.
        seed
            random number seed
        motif_length
            length of sequence units to consider. If not 1, length and start are
            converted (reduced) if necessary to be modulo motif_length
        moltype
            molecular type, can be string or instance

        Examples
        --------

        Create a sample alignment and an app that returns the first 4 positions
        of the alignment.

        >>> from cogent3 import make_aligned_seqs, get_app
        >>> aln = make_aligned_seqs({"s1": "GCAAGCGTTTAT", "s2": "GCTTTTGTCAAT"})
        >>> app_4 = get_app("fixed_length", length=4)
        >>> result = app_4(aln)
        >>> print(result.to_pretty())
        s1    GCAA
        s2    ..TT

        Return an alignment with ``length=4`` starting from the 2nd position.

        >>> app_4_start2 = get_app("fixed_length", length=4, start=2)
        >>> result = app_4_start2(aln)
        >>> print(result.to_pretty())
        s1    AAGC
        s2    TTTT

        The start position can be selected at random with ``random=True``. An
        optional ``seed`` can be provided to ensure the same start position is
        used when the app is called.

        >>> app_4_random = get_app("fixed_length", length=4, random=True, seed=1)
        >>> result = app_4_start2(aln)
        >>> print(result.to_pretty())
        s1    AAGC
        s2    TTTT

        Use ``motif_length=3`` to sample two triplets of ``length=6``.
        Sequences are split into non-overlapping sections of ``motif_length=3``
        before sampling (i.e. codon positions are preserved).

        >>> aln = make_aligned_seqs({"s1": "GCAAGCGTTTAT", "s2": "GCTTTTGTCAAT"})
        >>> app_motif3 = get_app("fixed_length", length=6, motif_length=3, random=True, seed=9)
        >>> result = app_motif3(aln)
        >>> print(result.to_pretty())
        s1    AGCTAT
        s2    TTTA..
        """
        diff = length % motif_length
        if diff != 0:
            length -= diff
        assert length % motif_length == 0

        self._length = length
        self._motif_length = motif_length
        if moltype:
            moltype = get_moltype(moltype)
        self._moltype = moltype
        if type(start) == str:
            assert start.lower().startswith("rand")
            random = False
        else:
            assert type(start) == int
            assert start >= 0
            diff = start % motif_length
            if diff != 0:
                start -= diff

        self._start = _GetStart(start)
        if seed:
            np_random.seed(seed)

        self._func = {False: self.truncated}.get(random, self.sample_positions)

    def truncated(self, aln):
        if self._moltype and self._moltype != aln.moltype:
            aln = aln.to_moltype(self._moltype)

        if len(aln) < self._length:
            msg = f"{len(aln)} < min_length {self._length}"
            return NotCompleted("FALSE", self.__class__.__name__, msg, source=aln)
        else:
            start = self._start(len(aln) - self._length)
            return aln[start : start + self._length]

    def sample_positions(self, aln):
        if self._moltype and self._moltype != aln.moltype:
            aln = aln.to_moltype(self._moltype)

        indices = array(range(len(aln)))
        if self._motif_length == 1:
            number = self._length
        else:
            number = self._length // self._motif_length

        pos = np_random.choice(
            indices.shape[0] // self._motif_length, number, replace=False
        )
        if self._motif_length == 1:
            result = indices[pos]
        else:
            dim = indices.shape[0] // self._motif_length
            indices.resize((dim, self._motif_length))
            result = indices[pos, :]

        result.sort(axis=0)
        result = aln.take_positions(result.flatten().tolist())
        return result

    T = Union[SerialisableType, AlignedSeqsType]

    def main(self, data: AlignedSeqsType) -> T:
        """return a fixed length alignment"""
        return self._func(data)


@define_app
class omit_bad_seqs:
    """Eliminates sequences from Alignment based on gap fraction, unique gaps."""

    def __init__(
        self,
        quantile: Optional[float] = None,
        gap_fraction: int = 1,
        moltype: MolTypes = "dna",
    ):
        """
        Parameters
        ----------
        quantile
            The number of gaps uniquely introduced in an alignment by each
            sequence are counted. The value corresponding to quantile is
            determined and all sequences whose unique gap count is larger than
            this cutoff are excluded. If None, this condition is not applied.
        gap_fraction
            sequences whose proportion of gaps is >= this value are excluded, the
            default excludes sequences that are just gaps.
        moltype
            molecular type, can be string or instance

        Examples
        --------

        Create a sample alignment and an app to remove sequences based on gap
        fraction. Use ``gap_fraction=0.5`` to omit sequences that contain 50%
        or more gaps.

        >>> from cogent3 import make_aligned_seqs, get_app
        >>> aln = make_aligned_seqs({
        ...     "s1": "---ACC---TT-",
        ...     "s2": "---ACC---TT-",
        ...     "s3": "---ACC---TT-",
        ...     "s4": "--AACCG-GTT-",
        ...     "s5": "--AACCGGGTTT",
        ...     "s6": "AGAACCGGGTT-",
        ...     "s7": "------------"
        ... }, moltype="dna")
        >>> app_frac_05 = get_app("omit_bad_seqs", gap_fraction=0.5)
        >>> result = app_frac_05(aln)
        >>> print(result.to_pretty())
        s4    --AACCG-GTT-
        s5    .......G...T
        s6    AG.....G....

        The ``quantile=0.8`` argument omits sequences that introduce gaps in the
        alignment. In the following example, sequence `s6` is omitted, as it
        uniquely introduces gaps in the first two positions of the alignment
        which exceeds the cutoff.

        >>> app = get_app("omit_bad_seqs", quantile=0.8)
        >>> result = app(aln)
        >>> print(result.to_pretty())
        s1    ---ACC---TT-
        s2    ............
        s3    ............
        s4    ..A...G.G...
        s5    ..A...GGG..T
        """
        if moltype:
            moltype = get_moltype(moltype)
        assert (
            moltype.label.lower() in "dna rna protein protein_with_stop"
        ), "moltype must be one of DNA, RNA or PROTEIN"
        self._quantile = quantile
        self._gap_fraction = gap_fraction
        self._moltype = moltype

    T = Union[SerialisableType, AlignedSeqsType]

    def main(self, aln: AlignedSeqsType) -> T:
        if self._moltype and self._moltype != aln.moltype:
            aln = aln.to_moltype(self._moltype)

        gaps_per_seq = aln.count_gaps_per_seq()
        length = len(aln)
        keep = [n for n, c in gaps_per_seq.items() if c / length < self._gap_fraction]
        result = aln.take_seqs(keep)
        if self._quantile is not None:
            result = result.omit_bad_seqs(quantile=self._quantile)
        return result


@define_app
class omit_duplicated:
    """Removes redundant sequences, recording dropped sequences in
    seqs.info.dropped."""

    def __init__(
        self,
        mask_degen: bool = False,
        choose: str = "longest",
        seed: Optional[int] = None,
        moltype: Optional[MolTypes] = None,
    ):
        """
        Parameters
        ----------
        mask_degen
            if True, degenerate characters are ignored
        choose
            choose a representative from sets of duplicated sequences.
            Valid values are None (all members of a duplicated set are excluded),
            'longest', 'random'.
        seed
            set random number seed. Only applied of choose=='random'
        moltype
            molecular type, can be string or instance

        Examples
        --------
        Removes redundant sequences from a sequence collection (aligned or
        unaligned).

        Create sample data with duplicated sequences.

        >>> from cogent3 import get_app, make_aligned_seqs
        >>> data = {
        ... "a": "ACGT",
        ... "b": "ACG-",  # identical to 'a' except has a gap
        ... "c": "ACGG",  # duplicate
        ... "d": "ACGG",  # duplicate
        ... "e": "AGTC"   # unique
        ... }

        Create an app that selects a representative of omits duplicate sequences.
        Setting ``choose="longest"`` selects the duplicated sequence with the least
        number of gaps and ambiguous characters. In this case, only one of 'c' and
        'd' will be retained.

        >>> seqs = make_aligned_seqs(data, moltype="DNA")
        >>> app = get_app("omit_duplicated", moltype="dna", choose="longest")
        >>> result = app(seqs)
        >>> print(result.to_pretty())
        a    ACGT
        b    ...-
        d    ...G
        e    .GTC

        Setting ``choose=None`` means only unique sequences are retained.

        >>> app = get_app("omit_duplicated", moltype="dna", choose=None)
        >>> result = app(seqs)
        >>> print(result.to_pretty())
        a    ACGT
        b    ...-
        e    .GTC

        Use the ``mask_degen`` argument to specify how to treat matches between
        sequences with degenerate characters. We create sample data first that
        has a DNA ambiguity code.

        >>> data = make_aligned_seqs({
        ... "s1": "ATCG",
        ... "s2": "ATYG",  # matches s1 with ambiguity
        ... "s3": "GGTA",
        ... },
        ... moltype="DNA")
        >>> app_dna = get_app("omit_duplicated", mask_degen=True, choose="longest", moltype="DNA")
        >>> result = app_dna(data)
        >>> print(result.to_pretty())
        s1    ATCG
        s3    GGTA
        """
        assert not choose or choose in "longestrandom"
        if moltype:
            moltype = get_moltype(moltype)
        self._moltype = moltype
        if choose == "random" and seed:
            np_random.seed(seed)

        self._mask_degen = mask_degen
        if choose == "longest":
            self._func = self.choose_longest
        elif choose == "random":
            self._func = self.choose_random
        else:
            self._func = self.take_unique

    def choose_longest(self, seqs):
        if self._moltype and self._moltype != seqs.moltype:
            seqs = seqs.to_moltype(self._moltype)

        duplicates = seqs.get_identical_sets(mask_degen=self._mask_degen)
        lengths = seqs.get_lengths()
        excludes = []
        for group in duplicates:
            group_lengths = [(lengths[n], n) for n in group]
            group_lengths.sort(reverse=True)
            excludes.extend([n for l, n in group_lengths[1:]])

        seqs = seqs.take_seqs(excludes, negate=True)
        return seqs

    def choose_random(self, seqs):
        if self._moltype and self._moltype != seqs.moltype:
            seqs = seqs.to_moltype(self._moltype)

        duplicates = seqs.get_identical_sets(mask_degen=self._mask_degen)
        excludes = []
        for group in duplicates:
            chosen = np_random.choice(list(group))
            group.remove(chosen)
            excludes.extend(group)

        seqs = seqs.take_seqs(excludes, negate=True)
        return seqs

    def take_unique(self, seqs):
        if self._moltype and self._moltype != seqs.moltype:
            seqs = seqs.to_moltype(self._moltype)

        duplicates = seqs.get_identical_sets(mask_degen=self._mask_degen)
        names = set()
        for dupes in duplicates:
            names.update(dupes)
        seqs = seqs.take_seqs(names, negate=True)
        return seqs

    T = Union[SerialisableType, SeqsCollectionType]

    def main(self, seqs: SeqsCollectionType) -> T:
        return self._func(seqs)


@define_app
class trim_stop_codons:
    """Removes terminal stop codons."""

    def __init__(self, gc: Union[str, int] = 1):
        """
        Parameters
        ----------
        gc
            identifier for a genetic code or a genetic code instance, defaults
            to standard genetic code. See https://cogent3.org/doc/cookbook/what_codes.html

        Returns
        -------
        A new sequence collection, or False if not all the named sequences are
        in the collection.

        Examples
        --------

        Removes trailing (terminal) stop codons from sequences in a sequence
        collection or alignment.

        Create a sample unaligned sequence data and an app to remove trailing
        stop codons.

        >>> from cogent3 import make_unaligned_seqs, make_aligned_seqs, get_app
        >>> ualn = make_unaligned_seqs(data={"s1": "AAATTTCCC", "s2": "AAATTTTAA"}, moltype="dna")
        >>> app = get_app("trim_stop_codons")
        >>> result = app(ualn)
        >>> print(result.to_dict())
        {'s1': 'AAATTTCCC', 's2': 'AAATTT'}

        Remove trailing stop codons from aligned sequences, using the genetic
        code for "Vertebrate Mitochondrial".

        For a list of all genetic codes, see
        https://cogent3.org/doc/cookbook/what_codes.html.

        >>> aln = make_aligned_seqs(data={"s1": "AAATTTCCC", "s2": "AAATTTTAA"}, moltype="dna")
        >>> app = get_app("trim_stop_codons", gc=2)
        >>> result = app(aln)
        >>> print(result.to_dict())
        {'s1': 'AAATTTCCC', 's2': 'AAATTT---'}
        """
        self._gc = gc

    T = Union[SerialisableType, SeqsCollectionType]

    def main(self, data: SeqsCollectionType) -> T:
        data = data.trim_stop_codons(gc=self._gc)
        return data
